#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
This package provides the necessary constraints for gas phase properties for
the oxidation of a chemical looping oxygen carrier.
Components - Oxygen (O2), Nitrogen (N2), Carbon Dioxide (CO2), Water (H2O)

Equations written in this model were derived from:
(1) B.E. Poling, J.M. Prausnitz, J.P. O'connell, The Properties of Gases and
Liquids, Mcgraw-Hill, New York, 2001.
(2) National Institute of Standards and Technology, NIST Chemistry WebBook,
https://webbook.nist.gov/chemistry/ (accessed March 10, 2018).

"""

# Import Pyomo libraries
from pyomo.environ import (Constraint,
                           Param,
                           PositiveReals,
                           Reals,
                           value,
                           Var,
                           units as pyunits)
from pyomo.util.calc_var_value import calculate_variable_from_constraint
# from pyomo.opt import SolverFactory

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        PhysicalParameterBlock,
                        StateBlockData,
                        StateBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        Component,
                        VaporPhase)
from idaes.core.util.initialization import (fix_state_vars,
                                            revert_state_vars,
                                            solve_indexed_blocks)
from idaes.core.util.misc import add_object_reference
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_unfixed_variables_in_activated_equalities)
import idaes.logger as idaeslog

# Some more information about this module
__author__ = "Chinedu Okoli"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("GasPhaseParameterBlock")
class PhysicalParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class
    Contains parameters and indexing sets associated with properties for
    oxidation of oxygen carrier with oxygen.
    """

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(PhysicalParameterData, self).build()

        self._state_block_class = GasPhaseStateBlock

        # Create Phase object
        self.Vap = VaporPhase()

        # Create Component objects
        self.N2 = Component()
        self.O2 = Component()
        self.CO2 = Component()
        self.H2O = Component()

        # Gas Constant
        self.gas_const = Param(within=PositiveReals,
                               default=8.314459848e-3,
                               doc='Gas Constant [kJ/mol.K]',
                               units=pyunits.kJ/pyunits.mol/pyunits.K)

        # Mol. weights of gas - units = kg/mol. ref: NIST webbook
        mw_comp_dict = {'O2': 0.032, 'N2': 0.028, 'CO2': 0.044, 'H2O': 0.018}
        self.mw_comp = Param(
                self.component_list,
                mutable=False,
                initialize=mw_comp_dict,
                doc="Molecular weights of gas components [kg/mol]",
                units=pyunits.kg/pyunits.mol)

        # Std. heat of formation of comp. - units = kJ/(mol comp) - ref: NIST
        enth_mol_form_comp_dict = {'O2': 0, 'N2': 0, 'CO2': -393.5224,
                                   'H2O': -241.8264}
        self.enth_mol_form_comp = Param(
                self.component_list,
                mutable=False,
                initialize=enth_mol_form_comp_dict,
                doc="Component molar heats of formation [kJ/mol]",
                units=pyunits.kJ/pyunits.mol)

        # Ideal gas spec. heat capacity parameters(Shomate) of
        # components - ref: NIST webbook. Shomate equations from NIST.
        # Parameters A-E are used for cp calcs while A-H are used for enthalpy
        # calc.
        # 1e3*cp_comp = A + B*T + C*T^2 + D*T^3 + E/(T^2)
        # where T = Temperature (K)/1000, and cp_comp = (kJ/mol.K)
        # H_comp = H - H(298.15) = A*T + B*T^2/2 + C*T^3/3 +
        # D*T^4/4 - E/T + F - H where T = Temp (K)/1000 and H_comp = (kJ/mol)
        cp_param_dict = {
                        ('O2', 1): 30.03235,
                        ('O2', 2): 8.772972,
                        ('O2', 3): -3.988133,
                        ('O2', 4): 0.788313,
                        ('O2', 5): -0.741599,
                        ('O2', 6): -11.32468,
                        ('O2', 7): 236.1663,
                        ('O2', 8): 0.0000,
                        ('N2', 1): 19.50583,
                        ('N2', 2): 19.88705,
                        ('N2', 3): -8.598535,
                        ('N2', 4): 1.369784,
                        ('N2', 5): 0.527601,
                        ('N2', 6): -4.935202,
                        ('N2', 7): 212.3900,
                        ('N2', 8): 0.0000,
                        ('CO2', 1): 24.9973500,
                        ('CO2', 2): 55.1869600,
                        ('CO2', 3): -33.6913700,
                        ('CO2', 4): 7.9483870,
                        ('CO2', 5): -0.1366380,
                        ('CO2', 6): -403.6075000,
                        ('CO2', 7): 228.2431000,
                        ('CO2', 8): -393.5224000,
                        ('H2O', 1): 30.0920000,
                        ('H2O', 2): 6.8325140,
                        ('H2O', 3): 6.7934350,
                        ('H2O', 4): -2.5344800,
                        ('H2O', 5): 0.0821390,
                        ('H2O', 6): -250.8810000,
                        ('H2O', 7): 223.3967000,
                        ('H2O', 8): -241.8264000
                        }
        self.cp_param = Param(self.component_list,
                              range(1, 10),
                              mutable=False,
                              initialize=cp_param_dict,
                              doc="Shomate equation heat capacity parameters")

        # Viscosity constants:
        # Reference: Perry and Green Handbook; McGraw Hill, 2008
        visc_d_param_dict = {('O2', 1): 1.101e-6, ('O2', 2): 0.5634,
                             ('O2', 3): 96.3, ('O2', 4): 0,
                             ('N2', 1): 6.5592e-7, ('N2', 2): 0.6081,
                             ('N2', 3): 54.714, ('N2', 4): 0,
                             ('CO2', 1): 2.148e-6, ('CO2', 2): 0.46,
                             ('CO2', 3): 290, ('CO2', 4): 0,
                             ('H2O', 1): 1.7096e-8, ('H2O', 2): 1.1146,
                             ('H2O', 3): 0, ('H2O', 4): 0}
        self.visc_d_param = Param(self.component_list,
                                  range(1, 10),
                                  mutable=True,
                                  initialize=visc_d_param_dict,
                                  doc="Dynamic viscosity constants")

        # Thermal conductivity constants:
        # Reference: Perry and Green Handbook; McGraw Hill, 2008
        therm_cond_param_dict = {('N2', 1): 3.3143e-4, ('N2', 2): 0.7722,
                                 ('N2', 3): 16.323, ('N2', 4): 0,
                                 ('O2', 1): 4.4994e-4, ('O2', 2): 0.7456,
                                 ('O2', 3): 56.699, ('O2', 4): 0,
                                 ('CO2', 1): 3.69, ('CO2', 2): -0.3838,
                                 ('CO2', 3): 964, ('CO2', 4): 1.86e6,
                                 ('H2O', 1): 6.204e-6, ('H2O', 2): 1.3973,
                                 ('H2O', 3): 0, ('H2O', 4): 0}
        self.therm_cond_param = Param(self.component_list,
                                      range(1, 10),
                                      mutable=True,
                                      initialize=therm_cond_param_dict,
                                      doc="Thermal conductivity constants")

        # Component diffusion volumes:
        # Ref: (1) Prop gas & liquids (2) Fuller et al. IECR, 58(5), 19, 1966
        diff_vol_param_dict = {'O2': 16.6, 'N2': 17.9,
                               'CO2': 26.9, 'H2O': 13.1}
        self.diff_vol_param = Param(self.component_list,
                                    mutable=True,
                                    initialize=diff_vol_param_dict,
                                    doc="Component diffusion volumes")

        self._eps = Param(initialize=1e-8,
                          doc="Smooth abs reformulation parameter")

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({
                'flow_mol': {'method': None, 'units': 'mol/s'},
                'pressure': {'method': None, 'units': 'bar'},
                'temperature': {'method': None, 'units': 'K'},
                'mole_frac_comp': {'method': None, 'units': None},
                'mw': {'method': '_mw', 'units': 'kg/mol'},
                'cp_mol': {'method': '_cp_mol', 'units': 'kJ/mol.K'},
                'cp_mol_comp': {'method': '_cp_mol_comp',
                                'units': 'kJ/mol.K'},
                'cp_mass': {'method': '_cp_mass', 'units': 'kJ/kg.K'},
                'dens_mol': {'method': '_dens_mol',
                             'units': 'mol/m^3'},
                'dens_mol_comp': {'method': '_dens_mol_comp',
                                  'units': 'mol/m^3'},
                'dens_mass': {'method': '_dens_mass',
                              'units': 'kg/m^3'},
                'enth_mol': {'method': '_enth_mol', 'units': 'kJ/mol'},
                'enth_mol_comp': {'method': '_enth_mol_comp',
                                  'units': 'kJ/mol'},
                'visc_d': {'method': '_visc_d', 'units': 'kg/m.s'},
                'therm_cond': {'method': '_therm_cond', 'units': 'kJ/m.K.s'},
                'diffusion_comp': {'method': '_diffusion_comp',
                                   'units': 'cm2/s'}})

        obj.add_default_units({'time': pyunits.s,
                               'length': pyunits.m,
                               'mass': pyunits.kg,
                               'amount': pyunits.mol,
                               'temperature': pyunits.K})
        # def add_default_units(self, u): u (dict): Key=property, Value=units
        # def add_properties(self, p): p (dict): Key=property, Value=PropertyMetadata or equiv. dict
        # def get_derived_units(self, units):
        # obj.get_derived_units("power") = pyunits.kJ * pyunits.s ** -1


class _GasPhaseStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to State Blocks as a
    whole, rather than individual elements of indexed State Blocks.
    """
    def initialize(blk, state_args=None, hold_state=False,
                   state_vars_fixed=False, outlvl=idaeslog.NOTSET,
                   solver="ipopt", optarg={"tol": 1e-8}):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.
                         Keys for the state_args dictionary are:
                         flow_mol, temperature, pressure and mole_frac_comp
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = "ipopt")
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="properties")

        init_log.info_high('Starting initialization')

        # Deactivate the constraints specific for non-inlet blocks i.e.
        # when defined state is False
        for k in blk.keys():
            if blk[k].config.defined_state is False:
                blk[k].sum_component_eqn.deactivate()

        # Fix state variables if not already fixed
        if state_vars_fixed is False:
            flags = fix_state_vars(blk, state_args)
        else:
            # Check when the state vars are fixed already result in dof 0
            for k in blk.keys():
                if degrees_of_freedom(blk[k]) != 0:
                    raise Exception("State vars fixed but degrees of freedom "
                                    "for state block is not zero during "
                                    "initialization.")

        # ---------------------------------------------------------------------
        # Initialise values
        for k in blk.keys():

            if hasattr(blk[k], "mw_eqn"):
                calculate_variable_from_constraint(
                            blk[k].mw,
                            blk[k].mw_eqn)

            if hasattr(blk[k], "ideal_gas"):
                calculate_variable_from_constraint(
                            blk[k].dens_mol,
                            blk[k].ideal_gas)

            if hasattr(blk[k], "dens_mass_basis"):
                calculate_variable_from_constraint(
                            blk[k].dens_mass,
                            blk[k].dens_mass_basis)

            if hasattr(blk[k], "mixture_heat_capacity_eqn"):
                calculate_variable_from_constraint(
                            blk[k].cp_mol,
                            blk[k].mixture_heat_capacity_eqn)

            if hasattr(blk[k], "cp_mass_basis"):
                calculate_variable_from_constraint(
                            blk[k].cp_mass,
                            blk[k].cp_mass_basis)

            if hasattr(blk[k], "visc_d_constraint"):
                calculate_variable_from_constraint(
                            blk[k].visc_d,
                            blk[k].visc_d_constraint)

            if hasattr(blk[k], "therm_cond_constraint"):
                calculate_variable_from_constraint(
                            blk[k].therm_cond,
                            blk[k].therm_cond_constraint)

            if hasattr(blk[k], "mixture_enthalpy_eqn"):
                calculate_variable_from_constraint(
                            blk[k].enth_mol,
                            blk[k].mixture_enthalpy_eqn)

            for j in blk[k]._params.component_list:

                if hasattr(blk[k], "comp_conc_eqn"):
                    calculate_variable_from_constraint(
                                blk[k].dens_mol_comp[j],
                                blk[k].comp_conc_eqn[j])

                if hasattr(blk[k], "diffusion_comp_constraint"):
                    calculate_variable_from_constraint(
                                blk[k].diffusion_comp[j],
                                blk[k].diffusion_comp_constraint[j])

                if hasattr(blk[k], "cp_shomate_eqn"):
                    calculate_variable_from_constraint(
                        blk[k].cp_mol_comp[j],
                        blk[k].cp_shomate_eqn[j])

                if hasattr(blk[k], "enthalpy_shomate_eqn"):
                    calculate_variable_from_constraint(
                            blk[k].enth_mol_comp[j],
                            blk[k].enthalpy_shomate_eqn[j])

        # Solve property block if non-empty
        free_vars = 0
        for k in blk.keys():
            free_vars += number_unfixed_variables_in_activated_equalities(
                blk[k])

        if free_vars > 0:
            # Create solver
            opt = get_solver(solver, optarg)
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solve_indexed_blocks(opt, [blk], tee=slc.tee)
        else:
            res = ""
        init_log.info_high("Initialization complete {}.".format(
            idaeslog.condition(res))
        )

        # ---------------------------------------------------------------------
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)

    def release_state(blk, flags, outlvl=0):
        """
        Method to relase state variables fixed during initialization.
        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of logging
        """
        if flags is None:
            return

        # Unfix state variables
        revert_state_vars(blk, flags)

        # Activate state variable related constraints
        for k in blk.keys():
            if blk[k].config.defined_state is False:
                blk[k].sum_component_eqn.activate()

        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="properties")
        init_log.info_high('States released.')


@declare_process_block_class("GasPhaseStateBlock",
                             block_class=_GasPhaseStateBlock)
class GasPhaseStateBlockData(StateBlockData):
    """
    Property package for gas phase properties of methane combustion in CLC FR
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(GasPhaseStateBlockData, self).build()

        # Object reference for molecular weight if needed by CV1D
        # Molecular weights
        add_object_reference(self, "mw_comp",
                             self.config.parameters.mw_comp)

        """List the necessary state variable objects."""
        self.flow_mol = Var(initialize=1.0,
                            domain=Reals,
                            doc='Component molar flowrate [mol/s]',
                            units=pyunits.mol/pyunits.s)
        self.mole_frac_comp = Var(
                self._params.component_list,
                domain=Reals,
                initialize=1 / len(self._params.component_list),
                doc='State component mole fractions [-]',
                units=pyunits.mol/pyunits.mol)
        self.pressure = Var(initialize=1.01325,
                            domain=Reals,
                            doc='State pressure [bar]',
                            units=pyunits.bar)
        self.temperature = Var(initialize=298.15,
                               domain=Reals,
                               doc='State temperature [K]',
                               units=pyunits.K)

        # Create standard constraints
        # Sum mole fractions if not inlet block
        if self.config.defined_state is False:
            def sum_component_eqn(b):
                return 1e2 == 1e2 * sum(b.mole_frac_comp[j]
                                        for j in b._params.component_list)
            self.sum_component_eqn = Constraint(rule=sum_component_eqn)

    def _mw(self):
        # Molecular weight of gas mixture
        self.mw = Var(domain=Reals,
                      initialize=1.0,
                      doc="Molecular weight of gas mixture [kg/mol]",
                      units=pyunits.kg/pyunits.mol)

        def mw_eqn(b):
            return (b.mw ==
                    sum(b.mole_frac_comp[j]*b._params.mw_comp[j]
                        for j in b._params.component_list))
        try:
            # Try to build constraint
            self.mw_eqn = Constraint(rule=mw_eqn)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.mw)
            self.del_component(self.mw_eqn)
            raise

    def _dens_mol(self):
        # Molar density
        self.dens_mol = Var(domain=Reals,
                            initialize=1.0,
                            doc="Molar density/concentration [mol/m3]",
                            units=pyunits.mol/pyunits.m**3)

        def ideal_gas(b):
            return (b.dens_mol*b._params.gas_const*b.temperature*1e-2 ==
                    b.pressure)
        try:
            # Try to build constraint
            self.ideal_gas = Constraint(rule=ideal_gas)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.dens_mol)
            self.del_component(self.ideal_gas)
            raise

    def _dens_mol_comp(self):
        # Mixture heat capacities
        self.dens_mol_comp = Var(self._params.component_list,
                                 domain=Reals,
                                 initialize=1.0,
                                 doc='Component molar concentration'
                                 '[mol/m3]',
                                 units=pyunits.mol/pyunits.m**3)

        def comp_conc_eqn(b, j):
            return (b.dens_mol_comp[j] ==
                    b.dens_mol*b.mole_frac_comp[j])
        try:
            # Try to build constraint
            self.comp_conc_eqn = Constraint(self._params.component_list,
                                            rule=comp_conc_eqn)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.dens_mol_comp)
            self.del_component(self.comp_conc_eqn)
            raise

    def _dens_mass(self):
        # Mass density
        self.dens_mass = Var(domain=Reals,
                             initialize=1.0,
                             doc="Mass density [kg/m3]",
                             units=pyunits.kg/pyunits.m**3)

        def dens_mass_basis(b):
            return b.dens_mass == b.mw*b.dens_mol
        try:
            # Try to build constraint
            self.dens_mass_basis = Constraint(rule=dens_mass_basis)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.dens_mass)
            self.del_component(self.dens_mass_basis)
            raise

    def _visc_d(self):
        # Mixture dynamic viscosity
        self.visc_d = Var(domain=Reals,
                          initialize=1e-5,
                          doc="Mixture dynamic viscosity [kg/m.s]",
                          units=pyunits.kg/pyunits.m/pyunits.s)

        def visc_d_comp(i):
            return self._params.visc_d_param[i, 1] * \
                    (self.temperature**self._params.visc_d_param[i, 2]) \
                    / ((1 + (self._params.visc_d_param[i, 3]/self.temperature))
                        + (self._params.visc_d_param[i, 4] /
                           (self.temperature**2)))

        def visc_d_constraint(b):
            return 1e6*b.visc_d == 1e6*sum(b.mole_frac_comp[i]*visc_d_comp(i)
                                           / (sum(b.mole_frac_comp[j]
                                                  * (b._params.mw_comp[j] /
                                                     b._params.mw_comp[i])**0.5
                                                  for j in
                                                  b._params.component_list))
                                           for i in
                                           b._params.component_list)
        try:
            # Try to build constraint
            self.visc_d_constraint = Constraint(rule=visc_d_constraint)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.visc_d)
            self.del_component(self.visc_d_constraint)
            raise

    def _diffusion_comp(self):
        # Component diffusion in a gas mixture - units of cm2/s to help scaling
        self.diffusion_comp = Var(self._params.component_list,
                                  domain=Reals,
                                  initialize=1e-5,
                                  doc='Component diffusion in a gas mixture'
                                  '[cm2/s]',
                                  units=pyunits.cm**2/pyunits.s)

        def D_bin(i, j):
            # 1e3 used to multiply MW to convert from kg/mol to kg/kmol
            return ((1.43e-3*(self.temperature**1.75) *
                     ((1e3 * self._params.mw_comp[i] +
                       1e3 * self._params.mw_comp[j])
                     / (2 * (1e3 * self._params.mw_comp[i]) *
                        (1e3*self._params.mw_comp[j])))**0.5)
                    / ((self.pressure)
                        * ((self._params.diff_vol_param[i]**(1/3))
                            + (self._params.diff_vol_param[j]**(1/3)))**2))

        def diffusion_comp_constraint(b, i):
            return (b.diffusion_comp[i]
                    * sum(b.mole_frac_comp[j]/D_bin(i, j)
                    for j in b._params.component_list if i != j)
                    == (1-b.mole_frac_comp[i]))
        try:
            # Try to build constraint
            self.diffusion_comp_constraint \
                = Constraint(self._params.component_list,
                             rule=diffusion_comp_constraint)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.diffusion_comp)
            self.del_component(self.diffusion_comp_constraint)
            raise

    def _therm_cond(self):
        # Thermal conductivity of gas
        self.therm_cond = Var(domain=Reals,
                              initialize=1e-5,
                              doc="Thermal conductivity of gas [kJ/m.K.s]",
                              units=pyunits.kJ/pyunits.m/pyunits.K/pyunits.s)

        def therm_cond_comp(i):
            return self._params.therm_cond_param[i, 1] \
                    * (self.temperature**self._params.therm_cond_param[i, 2]) \
                    / ((1 + (self._params.therm_cond_param[i, 3] /
                             self.temperature))
                        + (self._params.therm_cond_param[i, 4] /
                       (self.temperature**2)))

        def A_bin(i, j):
            return (1 + ((therm_cond_comp(j)/therm_cond_comp(i))**0.5)
                    * ((self._params.mw_comp[j] /
                        self._params.mw_comp[i])**0.25))**2 \
                    / (8*(1+(self._params.mw_comp[j] /
                             self._params.mw_comp[i])))**0.5

        def therm_cond_constraint(b):
            # The 1e-3 term is used as a conversion factor to a kJ basis
            return 1e6*b.therm_cond == 1e6*(1e-3) * \
                                       sum(b.mole_frac_comp[i]
                                           * therm_cond_comp(i)
                                           / (sum(b.mole_frac_comp[j] *
                                                  A_bin(i, j)**0.5
                                                  for j in
                                                  b._params.component_list))
                                           for i in
                                           b._params.component_list)
        try:
            # Try to build constraint
            self.therm_cond_constraint = Constraint(rule=therm_cond_constraint)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.therm_cond)
            self.del_component(self.therm_cond_constraint)
            raise

    def _cp_mol_comp(self):
        # Pure component vapour heat capacities
        self.cp_mol_comp = Var(self._params.component_list,
                               domain=Reals,
                               initialize=1.0,
                               doc="Pure component vapour heat capacities "
                               "[kJ/mol.K]",
                               units=pyunits.kJ/pyunits.mol/pyunits.K)

        def pure_component_cp_mol(b, j):
            return b.cp_mol_comp[j] == 1e-3*(
                        b._params.cp_param[j, 1] +
                        b._params.cp_param[j, 2]*(b.temperature*1e-3) +
                        b._params.cp_param[j, 3]*(b.temperature*1e-3)**2 +
                        b._params.cp_param[j, 4]*(b.temperature*1e-3)**3 +
                        b._params.cp_param[j, 5]/((b.temperature*1e-3)**2))
        try:
            # Try to build constraint
            self.cp_shomate_eqn = Constraint(self._params.component_list,
                                             rule=pure_component_cp_mol)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mol_comp)
            self.del_component(self.cp_shomate_eqn)
            raise

    def _cp_mol(self):
        # Mixture heat capacities
        self.cp_mol = Var(domain=Reals,
                          initialize=1.0,
                          doc="Mixture heat capacity [kJ/mol.K]",
                          units=pyunits.kJ/pyunits.mol/pyunits.K)

        def cp_mol(b):
            return b.cp_mol == sum(b.cp_mol_comp[j]*b.mole_frac_comp[j]
                                   for j in b._params.component_list)
        try:
            # Try to build constraint
            self.mixture_heat_capacity_eqn = Constraint(
                                            rule=cp_mol)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mol)
            self.del_component(self.mixture_heat_capacity_eqn)
            raise

    def _cp_mass(self):
        # Mixture heat capacities
        self.cp_mass = Var(domain=Reals,
                           initialize=1.0,
                           doc="Mixture heat capacity, mass-basis [kJ/kg.K]",
                           units=pyunits.kJ/pyunits.kg/pyunits.K)

        def cp_mass(b):
            return b.cp_mass*b.mw == b.cp_mol
        try:
            # Try to build constraint
            self.cp_mass_basis = Constraint(rule=cp_mass)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mass)
            self.del_component(self.cp_mass_basis)
            raise

    def _enth_mol_comp(self):
        # Pure component vapour enthalpies
        self.enth_mol_comp = Var(
                self._params.component_list,
                domain=Reals,
                initialize=1.0,
                doc="Pure component enthalpies [kJ/mol]",
                units=pyunits.kJ/pyunits.mol)

        def pure_comp_enthalpy(b, j):
            return b.enth_mol_comp[j] == (
                    b._params.cp_param[j, 1]*(b.temperature*1e-3) +
                    b._params.cp_param[j, 2]*((b.temperature*1e-3)**2)/2 +
                    b._params.cp_param[j, 3]*((b.temperature*1e-3)**3)/3 +
                    b._params.cp_param[j, 4]*((b.temperature*1e-3)**4)/4 -
                    b._params.cp_param[j, 5]/(b.temperature*1e-3) +
                    b._params.cp_param[j, 6] -
                    b._params.cp_param[j, 8])
        try:
            # Try to build constraint
            self.enthalpy_shomate_eqn = Constraint(self._params.component_list,
                                                   rule=pure_comp_enthalpy)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.enth_mol_comp)
            self.del_component(self.enthalpy_shomate_eqn)
            raise

    def _enth_mol(self):
        # Mixture molar enthalpy
        self.enth_mol = Var(
                            domain=Reals,
                            initialize=1.0,
                            doc='Mixture specific enthalpy [kJ/mol]',
                            units=pyunits.kJ/pyunits.mol)
        try:
            # Try to build constraint
            self.mixture_enthalpy_eqn = Constraint(expr=(
                        self.enth_mol == sum(self.mole_frac_comp[j] *
                                             self.enth_mol_comp[j]
                                             for j in
                                             self._params.component_list)))
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.enth_mol)
            self.del_component(self.mixture_enthalpy_eqn)
            raise

    def get_material_flow_terms(b, p, j):
        return b.flow_mol*b.mole_frac_comp[j]

    def get_enthalpy_flow_terms(b, p):
        return b.flow_mol*b.enth_mol

    def get_material_density_terms(b, p, j):
        return b.dens_mol_comp[j]

    def get_energy_density_terms(b, p):
        return b.dens_mol*b.enth_mol

    def define_state_vars(b):
        return {"flow_mol": b.flow_mol,
                "temperature": b.temperature,
                "pressure": b.pressure,
                "mole_frac_comp": b.mole_frac_comp}

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def model_check(blk):
        """
        Model checks for property block
        """
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            _log.error('{} Temperature set below lower bound.'
                       .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            _log.error('{} Temperature set above upper bound.'
                       .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            _log.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            _log.error('{} Pressure set above upper bound.'.format(blk.name))

    def default_material_balance_type(blk):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(blk):
        return EnergyBalanceType.enthalpyTotal



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

from pandas import DataFrame
from collections import OrderedDict
from pyomo.environ import value
from pyomo.network import Arc, Port

from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

__author__ = "John Eslick, Andrew Lee"


def arcs_to_stream_dict(
    blk, additional=None, descend_into=True, sort=False, prepend=None, s=None):
    """
    Creates a stream dictionary from the Arcs in a model, using the Arc names as
    keys. This can be used to automate the creation of the streams dictionary
    needed for the ``create_stream_table_dataframe()`` and  ``stream_states_dict()``
    functions.

    Args:
        blk (pyomo.environ._BlockData): Pyomo model to search for Arcs
        additional (dict): Additional states to add to the stream dictionary,
            which aren't represented by arcs in blk, for example feed or
            product streams without Arcs attached or states internal to a unit
            model.
        descend_into (bool): If True, search subblocks for Arcs as well. The
            default is True.
        sort (bool): If True sort keys and return an OrderedDict
        prepend (str): Prepend a string to the arc name joined with a '.'.
            This can be useful to prevent conflicting names when sub blocks
            contain Arcs that have the same names when used in combination
            with descend_into=False.
        s (dict): Add streams to an existing stream dict.

    Returns:
        Dictionary with Arc names as keys and the Arcs as values.

    """
    if s is None:
        s = {}
    for c in blk.component_objects(Arc, descend_into=descend_into):
        key = c.getname()
        if prepend is not None:
            key = ".".join([prepend, key])
        s[key] = c
    if additional is not None:
        s.update(additional)
    if sort:
        s = OrderedDict(sorted(s.items()))
    return s


def stream_states_dict(streams, time_point=0):
    """
    Method to create a dictionary of state block representing stream states.
    This takes a dict with stream name keys and stream values.

    Args:
        streams : dict with name keys and stream values. Names will be used as
            display names for stream table, and streams may be Arcs, Ports or
            StateBlocks.
        time_point : point in the time domain at which to generate stream table
            (default = 0)

    Returns:
        A pandas DataFrame containing the stream table data.
    """
    stream_dict = OrderedDict()
    def _stream_dict_add(sb, n, i=None):
        """add a line to the stream table"""
        if i is None:
            key = n
        else:
            key = "{}[{}]".format(n, i)
        stream_dict[key] = sb

    for n in streams.keys():
        try:
            if isinstance(streams[n], Arc):
                for i, a in streams[n].items():
                    sb = _get_state_from_port(a.ports[1], time_point)
                    _stream_dict_add(sb, n, i)
            elif isinstance(streams[n], Port):
                sb = _get_state_from_port(streams[n], time_point)
                _stream_dict_add(sb, n)
            else:
                sb = streams[n][time_point]
                _stream_dict_add(sb, n)
        except (AttributeError, KeyError):
            raise TypeError(
                f"Unrecognised component type for stream argument {streams[n]}."
                f" The stream_states_dict function only supports Arcs, "
                f"Ports or StateBlocks."
            )
    return stream_dict


def tag_state_quantities(blocks, attributes, labels, exception=False):
    """ Take a stream states dictionary, and return a tag dictionary for stream
    quantities.  This takes a dictionary (blk) that has state block labels as
    keys and state blocks as values.  The attributes are a list of attributes to
    tag.  If an element of the attribute list is list-like, the fist element is
    the attribute and the remaining elements are indexes.  Lables provides a list
    of attribute lables to be used to create the tag.  Tags are blk_key + label
    for the attribute.

    Args:
        blocks (dict): Dictionary of state blocks.  The key is the block label to
            be used in the tag, and the value is a state block.
        attributes (list-like): A list of attriutes to tag.  It is okay if a
            particular attribute does not exist in a state bock.  This allows
            you to mix state blocks with differnt sets of attributes. If an
            attribute is indexed, the attribute can be specified as a list or
            tuple where the first element is the attribute and the remaining
            elements are indexes.
        labels (list-like): These are attribute lables.  The order corresponds to the
            attribute list.  They are used to create the tags.  Tags are in the
            form blk.key + label.
        exception (bool): If True, raise exceptions releated to invalid or
            missing indexes. If false missing or bad indexes are ignored and
            None is used for the table value.  Setting this to False allows
            tables where some state blocks have the same attributes with differnt
            indexing. (default is True)

    Return:
        (dict): Dictionary where the keys are tags and the values are model
            attributes, usually Pyomo component data objects.
    """

    tags={}
    if labels is None:
        lables = attributes
        for a in attributes:
            if isinstance(a, (tuple, list)):
                if len(a) == 2:
                    # in case there are multiple indexes and user gives tuple
                    label = f"{a[0]}[{a[1]}]"
                if len(a) > 2:
                    label = f"{a[0]}[{a[1:]}]"
                else:
                    label = a[0]

    for key, s in blocks.items():
        for i, a in enumerate(attributes):
            j = None
            if isinstance(a, (list, tuple)):
                # if a is list or tuple, the first element should be the
                # attribute and the remaining elements should be indexes.
                if len(a) == 2:
                    j = a[1] # catch user supplying list-like of indexes
                if len(a) > 2:
                    j = a[1:]
                #if len(a) == 1, we'll say that's fine here.  Don't know why you
                #would put the attribute in a list-like if not indexed, but I'll
                #allow it.
                a = a[0]
            v = getattr(s, a, None)
            if j is not None and v is not None:
                try:
                    v = v[j]
                except KeyError:
                    if not exception:
                        v = None
                    else:
                        _log.error(f"{j} is not a valid index of {a}")
                        raise KeyError(f"{j} is not a valid index of {a}")
            try:
                value(v, exception=False)
            except TypeError:
                if not exception:
                    v = None
                else:
                    _log.error(
                        f"Cannot calculate value of {a} (may be subscriptable)")
                    raise TypeError(
                        f"Cannot calculate value of {a} (may be subscriptable)")
            except ZeroDivisionError:
                pass # this one is okay
            if v is not None:
                tags[f"{key}{labels[i]}"] = v
    return tags


def create_stream_table_dataframe(
    streams, true_state=False, time_point=0, orient="columns"
):
    """
    Method to create a stream table in the form of a pandas dataframe. Method
    takes a dict with name keys and stream values. Use an OrderedDict to list
    the streams in a specific order, otherwise the dataframe can be sorted
    later.

    Args:
        streams : dict with name keys and stream values. Names will be used as
            display names for stream table, and streams may be Arcs, Ports or
            StateBlocks.
        true_state : indicated whether the stream table should contain the
            display variables define in the StateBlock (False, default) or the
            state variables (True).
        time_point : point in the time domain at which to generate stream table
            (default = 0)
        orient : orientation of stream table. Accepted values are 'columns'
            (default) where streams are displayed as columns, or 'index' where
            stream are displayed as rows.

    Returns:
        A pandas DataFrame containing the stream table data.
    """
    stream_attributes = OrderedDict()
    stream_states = stream_states_dict(streams=streams, time_point=time_point)
    full_keys = []  # List of all rows in dataframe to fill in missing data
    for key, sb in stream_states.items():
        stream_attributes[key] = {}
        if true_state:
            disp_dict = sb.define_state_vars()
        else:
            disp_dict = sb.define_display_vars()
        for k in disp_dict:
            for i in disp_dict[k]:
                if i is None:
                    stream_attributes[key][k] = value(disp_dict[k][i])
                    if k not in full_keys:
                        full_keys.append(k)
                else:
                    stream_attributes[key][f"{k} {i}"] = value(disp_dict[k][i])
                    if f"{k} {i}" not in full_keys:
                        full_keys.append(f"{k} {i}")

    # Check for missing rows in any stream, and fill with "-" if needed
    for k, v in stream_attributes.items():
        for r in full_keys:
            if r not in v.keys():
                # Missing row, fill with placeholder
                v[r] = "-"

    return DataFrame.from_dict(stream_attributes, orient=orient)


def stream_table_dataframe_to_string(stream_table, **kwargs):
    """
    Method to print a stream table from a dataframe. Method takes any argument
    understood by DataFrame.to_string
    """
    # Set some default values for keyword arguments
    na_rep = kwargs.pop("na_rep", "-")
    justify = kwargs.pop("justify", "center")
    float_format = kwargs.pop("float_format", lambda x: "{:#.5g}".format(x))

    # Print stream table
    return stream_table.to_string(
        na_rep=na_rep, justify=justify, float_format=float_format, **kwargs
    )


def _get_state_from_port(port, time_point):
    # Check port for _state_block attribute
    try:
        if len(port._state_block) == 1:
            return port._state_block[0][time_point]
        else:
            return port._state_block[0][time_point, port._state_block[1]]
    except AttributeError:
        # Port was not created by IDAES add_port methods. Return exception for
        # the user to fix.
        raise ConfigurationError(
            f"Port {port.name} does not have a _state_block attribute, "
            f"thus cannot determine StateBlock to use for collecting data."
            f" Please provide the associated StateBlock instead, or use "
            f"the IDAES add_port methods to create the Port."
        )


def generate_table(blocks, attributes, heading=None, exception=True):
    """
    Create a Pandas DataFrame that contains a list of user-defined attributes
    from a set of Blocks.

    Args:
        blocks (dict): A dictionary with name keys and BlockData objects for
            values. Any name can be associated with a block. Use an OrderedDict
            to show the blocks in a specific order, otherwise the dataframe can
            be sorted later.
        attributes (list or tuple of strings): Attributes to report from a
            Block, can be a Var, Param, or Expression. If an attribute doesn't
            exist or doesn't have a valid value, it will be treated as missing
            data.
        heading (list or tuple of srings): A list of strings that will be used
            as column headings. If None the attribute names will be used.
        exception (bool): If True, raise exceptions releated to invalid or
            missing indexes. If false missing or bad indexes are ignored and
            None is used for the table value.  Setting this to False allows
            tables where some state blocks have the same attributes with differnt
            indexing. (default is True)
    Returns:
        (DataFrame): A Pandas dataframe containing a data table
    """
    if heading is None:
        heading = attributes
    st = DataFrame(columns=heading)
    row = [None] * len(attributes)  # not a big deal but save time on realloc
    for key, s in blocks.items():
        for i, a in enumerate(attributes):
            j = None
            if isinstance(a, (list, tuple)):
                # if a is list or tuple, assume index supplied
                try:
                    assert len(a) > 1
                except AssertionError:
                    _log.error(f"An index must be supplided for attribute {a[0]}")
                    raise AssertionError(
                        f"An index must be supplided for attribute {a[0]}")
                j = a[1:]
                a = a[0]
            v = getattr(s, a, None)
            if j is not None and v is not None:
                try:
                    v = v[j]
                except KeyError:
                    if not exception:
                        v = None
                    else:
                        _log.error(f"{j} is not a valid index of {a}")
                        raise KeyError(f"{j} is not a valid index of {a}")
            try:
                v = value(v, exception=False)
            except TypeError:
                if not exception:
                    v = None
                else:
                    _log.error(
                        f"Cannot calculate value of {a} (may be subscriptable)")
                    raise TypeError(
                        f"Cannot calculate value of {a} (may be subscriptable)")
            except ZeroDivisionError:
                v = None
            row[i] = v
        st.loc[key] = row
    return st



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
This is an example natural gas fuel cell (NGFC) power plant without CCS.
It uses a reduced order model created by PNNL to calculate the performance
of the solid oxide fuel cell based on a set a flowsheet conditions. More
details can be found in the associated jupyter notebook.
"""

import os
from collections import OrderedDict

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.environ import units as pyunits
from pyomo.network import Arc
from pyomo.common.fileutils import this_file_dir
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.util.infeasible import log_infeasible_constraints

# IDAES Imports
from idaes.core import FlowsheetBlock
from idaes.core.util import copy_port_values
from idaes.core.util import model_serializer as ms
from idaes.core.util.misc import svg_tag
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.tables import create_stream_table_dataframe

import idaes.core.util.scaling as iscale

from idaes.generic_models.properties.core.generic.generic_property import (
    GenericParameterBlock)
from idaes.generic_models.properties.core.generic.generic_reaction import (
    GenericReactionParameterBlock)

from idaes.generic_models.unit_models import (
    Mixer,
    Heater,
    HeatExchanger,
    PressureChanger,
    GibbsReactor,
    StoichiometricReactor,
    Separator,
    Translator)
from idaes.generic_models.unit_models.heat_exchanger import \
    delta_temperature_underwood_callback
from idaes.generic_models.unit_models.pressure_changer import \
    ThermodynamicAssumption
from idaes.generic_models.unit_models.separator import SplittingType
from idaes.generic_models.unit_models.mixer import MomentumMixingType

from idaes.power_generation.properties.natural_gas_PR import get_prop, get_rxn
from idaes.power_generation.properties.NGFC.ROM.SOFC_ROM import \
    build_SOFC_ROM, initialize_SOFC_ROM

import logging


def build_power_island(m):
    # create property packages - 3 property packages and 1 reaction
    NG_config = get_prop(
        components=['H2', 'CO', "H2O", 'CO2', 'CH4', "C2H6", "C3H8", "C4H10",
                    'N2', 'O2', 'Ar'])
    m.fs.NG_props = GenericParameterBlock(default=NG_config)

    syn_config = get_prop(
        components=["H2", "CO", "H2O", "CO2", "CH4", "N2", "O2", "Ar"])
    m.fs.syn_props = GenericParameterBlock(default=syn_config)

    air_config = get_prop(
        components=['H2O', 'CO2', 'N2', 'O2', 'Ar'])
    m.fs.air_props = GenericParameterBlock(default=air_config)

    m.fs.rxn_props = GenericReactionParameterBlock(
        default=get_rxn(
            m.fs.syn_props, reactions=["h2_cmb", "co_cmb", "ch4_cmb"]))

    # build anode side units
    m.fs.anode_mix = Mixer(
        default={"inlet_list": ["feed", "recycle"],
                 "property_package": m.fs.NG_props})

    m.fs.anode_hx = HeatExchanger(
        default={"delta_temperature_callback":
                 delta_temperature_underwood_callback,
                 "shell": {"property_package": m.fs.syn_props,
                           "has_pressure_change": True},
                 "tube": {"property_package": m.fs.NG_props,
                          "has_pressure_change": True}})

    m.fs.prereformer = GibbsReactor(
        default={"has_heat_transfer": True,
                 "has_pressure_change": True,
                 "inert_species": ["N2", "Ar", "O2"],
                 "property_package": m.fs.NG_props})

    m.fs.anode_translator = Translator(
        default={"outlet_state_defined": True,
                 "inlet_property_package": m.fs.NG_props,
                 "outlet_property_package": m.fs.syn_props})

    # C2H6, C3H8, and C4H10 are not present in the system after the prereformer
    # the property package is changed to improve model robustness
    @m.fs.anode_translator.Constraint(m.fs.time)
    def anode_translator_F(b, t):
        return b.inlet.flow_mol[t] == b.outlet.flow_mol[t]

    @m.fs.anode_translator.Constraint(m.fs.time)
    def anode_translator_T(b, t):
        return b.inlet.temperature[t] == b.outlet.temperature[t]

    @m.fs.anode_translator.Constraint(m.fs.time)
    def anode_translator_P(b, t):
        return b.inlet.pressure[t] == b.outlet.pressure[t]

    @m.fs.anode_translator.Constraint(m.fs.time, m.fs.syn_props.component_list)
    def anode_translator_x(b, t, j):
        return b.inlet.mole_frac_comp[t, j] == b.outlet.mole_frac_comp[t, j]

    m.fs.fuel_cell_mix = Mixer(
        default={"inlet_list": ["fuel_inlet", "ion_inlet"],
                 "momentum_mixing_type": MomentumMixingType.none,
                 "property_package": m.fs.syn_props})

    # outlet pressure should be equal to fuel inlet pressure because oxygen is
    # traveling through a solid electrolyte
    @m.fs.fuel_cell_mix.Constraint(m.fs.time)
    def ion_mix_constraint(b, t):
        return b.outlet.pressure[t] == b.fuel_inlet.pressure[t]

    m.fs.anode = GibbsReactor(
        default={"has_heat_transfer": True,
                 "has_pressure_change": True,
                 "inert_species": ["N2", "Ar"],
                 "property_package": m.fs.syn_props})

    m.fs.anode_recycle = Separator(
        default={"outlet_list": ["exhaust", "recycle"],
                 "property_package": m.fs.syn_props})

    m.fs.anode_blower = PressureChanger(
        default={'compressor': True,
                 'property_package': m.fs.syn_props,
                 'thermodynamic_assumption':
                     ThermodynamicAssumption.isentropic})

    m.fs.recycle_translator = Translator(
        default={"outlet_state_defined": True,
                 "inlet_property_package": m.fs.syn_props,
                 "outlet_property_package": m.fs.NG_props})

    # recycle is fed back to a part of the system that contains higher
    # hydrocarbons, so property package must be translated
    @m.fs.recycle_translator.Constraint(m.fs.time)
    def recycle_translator_F(b, t):
        return b.inlet.flow_mol[t] == b.outlet.flow_mol[t]

    @m.fs.recycle_translator.Constraint(m.fs.time)
    def recycle_translator_T(b, t):
        return b.inlet.temperature[t] == b.outlet.temperature[t]

    @m.fs.recycle_translator.Constraint(m.fs.time)
    def recycle_translator_P(b, t):
        return b.inlet.pressure[t] == b.outlet.pressure[t]

    @m.fs.recycle_translator.Constraint(m.fs.time,
                                        m.fs.syn_props.component_list)
    def recycle_translator_x(b, t, j):
        return b.inlet.mole_frac_comp[t, j] == b.outlet.mole_frac_comp[t, j]

    m.fs.recycle_translator.outlet.mole_frac_comp[0, 'C2H6'].fix(0)
    m.fs.recycle_translator.outlet.mole_frac_comp[0, 'C3H8'].fix(0)
    m.fs.recycle_translator.outlet.mole_frac_comp[0, 'C4H10'].fix(0)

    # build cathode side units
    m.fs.air_blower = PressureChanger(
        default={'compressor': True,
                 'property_package': m.fs.air_props,
                 'thermodynamic_assumption':
                     ThermodynamicAssumption.isentropic})

    m.fs.cathode_hx = HeatExchanger(
        default={"delta_temperature_callback":
                 delta_temperature_underwood_callback,
                 "shell": {"property_package": m.fs.air_props,
                           "has_pressure_change": True},
                 "tube": {"property_package": m.fs.air_props,
                          "has_pressure_change": True}})

    m.fs.cathode_mix = Mixer(
        default={"inlet_list": ["feed", "recycle"],
                 "property_package": m.fs.air_props})

    # represents oxygen moving across SOFC electrolyte
    m.fs.cathode = Separator(
        default={"outlet_list": ["air_outlet", "ion_outlet"],
                 "split_basis": SplittingType.componentFlow,
                 "property_package": m.fs.air_props})

    m.fs.cathode.split_fraction[0, 'ion_outlet', 'H2O'].fix(0)
    m.fs.cathode.split_fraction[0, 'ion_outlet', 'CO2'].fix(0)
    m.fs.cathode.split_fraction[0, 'ion_outlet', 'N2'].fix(0)
    m.fs.cathode.split_fraction[0, 'ion_outlet', 'Ar'].fix(0)

    m.fs.cathode_translator = Translator(
        default={"outlet_state_defined": True,
                 "inlet_property_package": m.fs.air_props,
                 "outlet_property_package": m.fs.syn_props})

    # cathode side O2 crosses the electrolyte to anode side
    @m.fs.cathode_translator.Constraint(m.fs.time)
    def cathode_translator_F(b, t):
        return b.inlet.flow_mol[t] == b.outlet.flow_mol[t]

    @m.fs.cathode_translator.Constraint(m.fs.time)
    def cathode_translator_T(b, t):
        return b.inlet.temperature[t] == b.outlet.temperature[t]

    @m.fs.cathode_translator.Constraint(m.fs.time)
    def cathode_translator_P(b, t):
        return b.inlet.pressure[t] == b.outlet.pressure[t]

    m.fs.cathode_translator.outlet.mole_frac_comp.fix(0)
    m.fs.cathode_translator.outlet.mole_frac_comp[0, 'O2'].fix(1)

    m.fs.cathode_heat = Heater(
        default={"has_pressure_change": True,
                 "property_package": m.fs.air_props})

    m.fs.cathode_recycle = Separator(
        default={"outlet_list": ["exhaust", "recycle"],
                 "property_package": m.fs.air_props})

    m.fs.cathode_blower = PressureChanger(
        default={'compressor': True,
                 'property_package': m.fs.air_props,
                 'thermodynamic_assumption':
                     ThermodynamicAssumption.isentropic})

    # build combustor and HRSG units
    m.fs.cathode_exhaust_split = Separator(
        default={"outlet_list": ["exhaust_outlet", "combustor_outlet"],
                 "property_package": m.fs.air_props})

    m.fs.cathode_expander = PressureChanger(
        default={"compressor": False,
                 "property_package": m.fs.air_props,
                 'thermodynamic_assumption':
                     ThermodynamicAssumption.isentropic})

    m.fs.cathode_HRSG = Heater(
        default={"has_pressure_change": True,
                 "property_package": m.fs.air_props})

    m.fs.cathode_exhaust_translator = Translator(
        default={"outlet_state_defined": True,
                 "inlet_property_package": m.fs.air_props,
                 "outlet_property_package": m.fs.syn_props})

    # for the portion of the cathode exhaust entering the combustor
    @m.fs.cathode_exhaust_translator.Constraint(m.fs.time)
    def cathode_exhaust_translator_F(b, t):
        return b.inlet.flow_mol[t] == b.outlet.flow_mol[t]

    @m.fs.cathode_exhaust_translator.Constraint(m.fs.time)
    def cathode_exhaust_translator_T(b, t):
        return b.inlet.temperature[t] == b.outlet.temperature[t]

    @m.fs.cathode_exhaust_translator.Constraint(m.fs.time)
    def cathode_exhaust_translator_P(b, t):
        return b.inlet.pressure[t] == b.outlet.pressure[t]

    @m.fs.cathode_exhaust_translator.Constraint(m.fs.time,
                                                m.fs.air_props.component_list)
    def cathode_exhaust_translator_x(b, t, j):
        return b.inlet.mole_frac_comp[t, j] == b.outlet.mole_frac_comp[t, j]

    for j in m.fs.syn_props.component_list:
        if j not in m.fs.air_props.component_list:
            m.fs.cathode_exhaust_translator.outlet.mole_frac_comp[0, j].fix(0)

    m.fs.combustor_mix = Mixer(
        default={"inlet_list": ["anode_inlet", "cathode_inlet"],
                 "property_package": m.fs.syn_props})

    m.fs.combustor = StoichiometricReactor(
        default={"has_heat_of_reaction": False,
                 "has_heat_transfer": False,
                 "has_pressure_change": True,
                 "property_package": m.fs.syn_props,
                 "reaction_package": m.fs.rxn_props})

    m.fs.combustor_expander = PressureChanger(
        default={"compressor": False,
                 "property_package": m.fs.syn_props,
                 'thermodynamic_assumption':
                     ThermodynamicAssumption.isentropic})

    m.fs.anode_HRSG = Heater(
        default={"has_pressure_change": True,
                 "property_package": m.fs.syn_props})

    # build arcs to connect unit operations
    # arcs for anode side
    m.fs.ANODE_MIXER = Arc(
        source=m.fs.anode_mix.outlet, destination=m.fs.anode_hx.tube_inlet)

    m.fs.PREREF_IN = Arc(
        source=m.fs.anode_hx.tube_outlet, destination=m.fs.prereformer.inlet)

    m.fs.PREREF_OUT = Arc(
        source=m.fs.prereformer.outlet,
        destination=m.fs.anode_translator.inlet)

    m.fs.ANODE_TRANS_OUT = Arc(
        source=m.fs.anode_translator.outlet,
        destination=m.fs.fuel_cell_mix.fuel_inlet)

    m.fs.ANODE_IN = Arc(
        source=m.fs.fuel_cell_mix.outlet, destination=m.fs.anode.inlet)

    m.fs.ANODE = Arc(
        source=m.fs.anode.outlet, destination=m.fs.anode_recycle.inlet)

    m.fs.ANODE_RECYCLE_HX = Arc(
        source=m.fs.anode_recycle.exhaust,
        destination=m.fs.anode_hx.shell_inlet)

    m.fs.ANODE_RECYCLE = Arc(
        source=m.fs.anode_recycle.recycle, destination=m.fs.anode_blower.inlet)

    m.fs.ANODE_BLOWER = Arc(
        source=m.fs.anode_blower.outlet,
        destination=m.fs.recycle_translator.inlet)

    m.fs.REC_TRANS_OUT = Arc(
        source=m.fs.recycle_translator.outlet,
        destination=m.fs.anode_mix.recycle)

    # arcs for cathode side
    m.fs.AIR_BLOWER = Arc(
        source=m.fs.air_blower.outlet, destination=m.fs.cathode_hx.tube_inlet)

    m.fs.CATHODE_HX_COLD = Arc(
        source=m.fs.cathode_hx.tube_outlet, destination=m.fs.cathode_mix.feed)

    m.fs.CATHODE_MIXER = Arc(
        source=m.fs.cathode_mix.outlet, destination=m.fs.cathode.inlet)

    m.fs.O2_IONS = Arc(
        source=m.fs.cathode.ion_outlet,
        destination=m.fs.cathode_translator.inlet)

    m.fs.O2_IONS_TRANSLATE = Arc(
        source=m.fs.cathode_translator.outlet,
        destination=m.fs.fuel_cell_mix.ion_inlet)

    m.fs.CATHODE = Arc(
        source=m.fs.cathode.air_outlet, destination=m.fs.cathode_heat.inlet)

    m.fs.CATHODE_HEAT = Arc(
        source=m.fs.cathode_heat.outlet,
        destination=m.fs.cathode_recycle.inlet)

    m.fs.CATHODE_RECYCLE_HX = Arc(
        source=m.fs.cathode_recycle.exhaust,
        destination=m.fs.cathode_hx.shell_inlet)

    m.fs.CATHODE_RECYCLE = Arc(
        source=m.fs.cathode_recycle.recycle,
        destination=m.fs.cathode_blower.inlet)

    m.fs.CATHODE_BLOWER = Arc(
        source=m.fs.cathode_blower.outlet,
        destination=m.fs.cathode_mix.recycle)

    # arcs for combustor and HRSG
    m.fs.CATHODE_HX_HOT_OUT = Arc(
        source=m.fs.cathode_hx.shell_outlet,
        destination=m.fs.cathode_exhaust_split.inlet)

    m.fs.CATHODE_EXPANDER_IN = Arc(
        source=m.fs.cathode_exhaust_split.exhaust_outlet,
        destination=m.fs.cathode_expander.inlet)

    m.fs.CATHODE_HRSG_IN = Arc(
        source=m.fs.cathode_expander.outlet,
        destination=m.fs.cathode_HRSG.inlet)

    m.fs.CATH_EXH_TRANS_IN = Arc(
        source=m.fs.cathode_exhaust_split.combustor_outlet,
        destination=m.fs.cathode_exhaust_translator.inlet)

    m.fs.CATH_EXH_TRANS_OUT = Arc(
        source=m.fs.cathode_exhaust_translator.outlet,
        destination=m.fs.combustor_mix.cathode_inlet)

    m.fs.ANODE_HX_HOT_OUT = Arc(
        source=m.fs.anode_hx.shell_outlet,
        destination=m.fs.combustor_mix.anode_inlet)

    m.fs.COMBUST_IN = Arc(
        source=m.fs.combustor_mix.outlet,
        destination=m.fs.combustor.inlet)

    m.fs.COMBUST_EXPANDER_IN = Arc(
        source=m.fs.combustor.outlet,
        destination=m.fs.combustor_expander.inlet)

    m.fs.ANODE_HRSG_IN = Arc(
        source=m.fs.combustor_expander.outlet,
        destination=m.fs.anode_HRSG.inlet)

    pyo.TransformationFactory("network.expand_arcs").apply_to(m.fs)


def set_power_island_inputs(m):
    ###################
    # anode side inputs
    ###################

    # syngas feed conditions
    m.fs.anode_mix.feed.flow_mol.fix(3639)  # mol/s
    m.fs.anode_mix.feed.temperature.fix(621.3)  # K
    m.fs.anode_mix.feed.pressure.fix(137895)  # Pa, equal to 20 psia
    m.fs.anode_mix.feed.mole_frac_comp[0, 'CH4'].fix(0.1787)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'C2H6'].fix(0.0061)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'C3H8'].fix(0.0013)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'C4H10'].fix(0.0008)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'CO'].fix(0.0909)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'CO2'].fix(0.0438)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'H2'].fix(0.2752)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'H2O'].fix(0.1118)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'N2'].fix(0.2879)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'O2'].fix(0.0000)
    m.fs.anode_mix.feed.mole_frac_comp[0, 'Ar'].fix(0.0034)

    # anode heat exchanger
    m.fs.anode_hx.tube.deltaP.fix(-137.9)  # equal to -0.02 psi
    m.fs.anode_hx.shell.deltaP.fix(-137.9)  # equal to -0.02 psi
    m.fs.anode_hx.area.fix(12664)  # m2
    m.fs.anode_hx.overall_heat_transfer_coefficient.fix(80)  # W/m2K

    # prereformer and anode
    m.fs.prereformer.heat_duty.fix(0)
    m.fs.prereformer.deltaP.fix(-137.9)  # equal to -0.02 psi

    m.fs.anode.outlet.temperature.fix(979.93)  # K, does not stay fixed
    m.fs.anode.outlet.pressure.fix(137137)  # Pa, equal to 19.89 psia

    # anode recycle and blower
    m.fs.anode_recycle.split_fraction[0, 'recycle'].fix(0.445)

    m.fs.anode_blower.outlet.pressure.fix(137888)  # Pa, equal to 20 psia
    m.fs.anode_blower.efficiency_isentropic.fix(0.8)

    #####################
    # cathode side inputs
    #####################

    # air feed conditions
    m.fs.air_blower.inlet.flow_mol.fix(17915.5)  # mol/s
    m.fs.air_blower.inlet.temperature.fix(288.15)  # K, equal to 59 F
    m.fs.air_blower.inlet.pressure.fix(101353)  # Pa, equal to 14.7 psia
    m.fs.air_blower.inlet.mole_frac_comp[0, 'H2O'].fix(0.0104)
    m.fs.air_blower.inlet.mole_frac_comp[0, 'CO2'].fix(0.0003)
    m.fs.air_blower.inlet.mole_frac_comp[0, 'N2'].fix(0.7722)
    m.fs.air_blower.inlet.mole_frac_comp[0, 'O2'].fix(0.2077)
    m.fs.air_blower.inlet.mole_frac_comp[0, 'Ar'].fix(0.0094)

    # air blower
    m.fs.air_blower.outlet.pressure.fix(111006)  # equal to 16.1 psi
    m.fs.air_blower.efficiency_isentropic.fix(0.82)

    # cathode heat exchanger
    m.fs.cathode_hx.tube_outlet.pressure.fix(105490)  # equal to 15.3 psia
    m.fs.cathode_hx.shell.deltaP.fix(-1379)  # equal to -0.2 psi
    m.fs.cathode_hx.area.fix(17159)  # m2
    m.fs.cathode_hx.overall_heat_transfer_coefficient.fix(80)  # W/m2K

    # cathode
    m.fs.cathode.ion_outlet.flow_mol.fix(1670)

    m.fs.cathode_heat.outlet.temperature.fix(999.4)  # K
    m.fs.cathode_heat.outlet.pressure.fix(104111)  # Pa, equal to 15.1 psi

    # cathode recycle and blower
    m.fs.cathode_recycle.split_fraction[0, 'recycle'].fix(0.5)

    m.fs.cathode_blower.outlet.pressure.fix(105490)  # Pa, equal to 15.3 psi
    m.fs.cathode_blower.efficiency_isentropic.fix(0.8)

    #####################
    # HRSG section inputs
    #####################

    m.fs.cathode_exhaust_split.split_fraction[0, 'combustor_outlet'].fix(0.3)

    m.fs.cathode_expander.deltaP.fix(-6.9)  # Pa, equal to -0.001 psi
    m.fs.cathode_expander.efficiency_isentropic.fix(0.95)

    m.fs.cathode_HRSG.outlet.temperature.fix(405.7)  # K, equal to 270 F
    m.fs.cathode_HRSG.deltaP.fix(-1379)  # Pa, equal to -0.2 psi

    m.fs.combustor.deltaP.fix(-6895)  # equal to -1 psi
    m.fs.combustor.outlet.mole_frac_comp[0, "H2"].fix(0)
    m.fs.combustor.outlet.mole_frac_comp[0, "CO"].fix(0)
    m.fs.combustor.outlet.mole_frac_comp[0, "CH4"].fix(0)

    m.fs.combustor_expander.deltaP.fix(-6.9)  # Pa, equal to -0.001 psi
    m.fs.combustor_expander.efficiency_isentropic.fix(0.95)

    m.fs.anode_HRSG.outlet.temperature.fix(405.7)  # K, equal to 270 F
    m.fs.anode_HRSG.deltaP.fix(-1379)  # Pa, equal to -0.2 psi


def build_reformer(m):
    # build unit models
    m.fs.reformer_recuperator = HeatExchanger(
        default={"delta_temperature_callback":
                 delta_temperature_underwood_callback,
                 "shell": {"property_package": m.fs.NG_props},
                 "tube": {"property_package": m.fs.NG_props}})

    m.fs.NG_expander = PressureChanger(
        default={'compressor': False,
                 'property_package': m.fs.NG_props,
                 'thermodynamic_assumption':
                     ThermodynamicAssumption.isentropic})

    m.fs.reformer_bypass = Separator(
        default={"outlet_list": ["reformer_outlet", "bypass_outlet"],
                 "property_package": m.fs.NG_props})

    m.fs.air_compressor_s1 = PressureChanger(
        default={"compressor": True,
                 "property_package": m.fs.NG_props,
                 "thermodynamic_assumption":
                     ThermodynamicAssumption.isentropic})

    # sets air flow to reformer based on NG flow
    @m.fs.Constraint()
    def reformer_air_rule(fs):
        ng_flow = m.fs.reformer_recuperator.tube_inlet.flow_mol[0]
        air_flow = m.fs.air_compressor_s1.inlet.flow_mol[0]
        IR = m.fs.reformer_bypass.split_fraction[0, 'bypass_outlet']
        return air_flow == 2.868*(1 - IR)*ng_flow

    m.fs.intercooler_s1 = Heater(
        default={"property_package": m.fs.NG_props,
                 "has_pressure_change": True})

    m.fs.air_compressor_s2 = PressureChanger(
        default={"compressor": True,
                 "property_package": m.fs.NG_props,
                 "thermodynamic_assumption":
                     ThermodynamicAssumption.isentropic})

    m.fs.intercooler_s2 = Heater(
        default={"property_package": m.fs.NG_props,
                 "has_pressure_change": True})

    m.fs.reformer_mix = Mixer(
        default={"inlet_list": ["gas_inlet", "oxygen_inlet", "steam_inlet"],
                 "property_package": m.fs.NG_props})

    # sets steam flow to reformer based on NG flow
    @m.fs.Constraint()
    def reformer_steam_flow(fs):
        ng_flow = m.fs.reformer_recuperator.tube_inlet.flow_mol[0]
        steam_flow = m.fs.reformer_mix.steam_inlet.flow_mol[0]
        IR = m.fs.reformer_bypass.split_fraction[0, 'bypass_outlet']
        return steam_flow == (1 - IR)*ng_flow

    m.fs.reformer = GibbsReactor(
        default={"has_heat_transfer": True,
                 "has_pressure_change": True,
                 "inert_species": ["N2", "Ar"],
                 "property_package": m.fs.NG_props})

    m.fs.bypass_rejoin = Mixer(
        default={"inlet_list": ["syngas_inlet", "bypass_inlet"],
                 "property_package": m.fs.NG_props})

    # build and connect arcs
    m.fs.TO_NG_EXP = Arc(
        source=m.fs.reformer_recuperator.tube_outlet,
        destination=m.fs.NG_expander.inlet)

    m.fs.NG_EXP_OUT = Arc(
        source=m.fs.NG_expander.outlet, destination=m.fs.reformer_bypass.inlet)

    m.fs.TO_REF = Arc(
        source=m.fs.reformer_bypass.reformer_outlet,
        destination=m.fs.reformer_mix.gas_inlet)

    m.fs.STAGE_1_OUT = Arc(
        source=m.fs.air_compressor_s1.outlet,
        destination=m.fs.intercooler_s1.inlet)

    m.fs.IC_1_OUT = Arc(
        source=m.fs.intercooler_s1.outlet,
        destination=m.fs.air_compressor_s2.inlet)

    m.fs.STAGE_2_OUT = Arc(
        source=m.fs.air_compressor_s2.outlet,
        destination=m.fs.intercooler_s2.inlet)

    m.fs.IC_2_OUT = Arc(
        source=m.fs.intercooler_s2.outlet,
        destination=m.fs.reformer_mix.oxygen_inlet)

    m.fs.REF_IN = Arc(
        source=m.fs.reformer_mix.outlet, destination=m.fs.reformer.inlet)

    m.fs.REF_OUT = Arc(
        source=m.fs.reformer.outlet,
        destination=m.fs.reformer_recuperator.shell_inlet)

    m.fs.REF_RECUP_OUT = Arc(
        source=m.fs.reformer_recuperator.shell_outlet,
        destination=m.fs.bypass_rejoin.syngas_inlet)

    m.fs.REF_BYPASS = Arc(
        source=m.fs.reformer_bypass.bypass_outlet,
        destination=m.fs.bypass_rejoin.bypass_inlet)

    pyo.TransformationFactory("network.expand_arcs").apply_to(m.fs)


def set_reformer_inputs(m):
    # natural gas feed conditions
    m.fs.reformer_recuperator.tube_inlet.flow_mol.fix(1161.9)  # mol/s
    m.fs.reformer_recuperator.tube_inlet.temperature.fix(288.15)  # K
    m.fs.reformer_recuperator.tube_inlet.pressure.fix(
        3447379)  # Pa, equal to 500 psia
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'CH4'].fix(0.931)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'C2H6'].fix(0.032)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'C3H8'].fix(0.007)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'C4H10'].fix(0.004)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'CO'].fix(0)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'CO2'].fix(0.01)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'H2'].fix(0)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'H2O'].fix(0)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'N2'].fix(0.016)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'O2'].fix(0)
    m.fs.reformer_recuperator.tube_inlet.mole_frac_comp[0, 'Ar'].fix(0)

    # recuperator conditions
    m.fs.reformer_recuperator.tube_outlet.temperature.fix(1010.93)
    m.fs.reformer_recuperator.overall_heat_transfer_coefficient.fix(80)

    # natural gas expander conditions
    m.fs.NG_expander.outlet.pressure.fix(206843)  # equal to 30 psia
    m.fs.NG_expander.efficiency_isentropic.fix(0.9)

    # internal reformation percentage
    m.fs.reformer_bypass.split_fraction[0, 'bypass_outlet'].fix(0.6)

    # air to reformer
    m.fs.air_compressor_s1.inlet.flow_mol[0] == 1332.9  # mol/s
    m.fs.air_compressor_s1.inlet.temperature.fix(288.15)  # K
    m.fs.air_compressor_s1.inlet.pressure.fix(101353)  # Pa, equal to 14.7 psia
    m.fs.air_compressor_s1.inlet.mole_frac_comp.fix(0)
    m.fs.air_compressor_s1.inlet.mole_frac_comp[0, 'CO2'].fix(0.0003)
    m.fs.air_compressor_s1.inlet.mole_frac_comp[0, 'H2O'].fix(0.0104)
    m.fs.air_compressor_s1.inlet.mole_frac_comp[0, 'N2'].fix(0.7722)
    m.fs.air_compressor_s1.inlet.mole_frac_comp[0, 'O2'].fix(0.2077)
    m.fs.air_compressor_s1.inlet.mole_frac_comp[0, 'Ar'].fix(0.0094)

    # air compressors and intercoolers
    m.fs.air_compressor_s1.outlet.pressure.fix(144790)  # Pa, equal to 21 psia
    m.fs.air_compressor_s1.efficiency_isentropic.fix(0.84)

    m.fs.intercooler_s1.outlet.temperature.fix(310.93)  # K, equal to 100 F
    m.fs.intercooler_s1.deltaP.fix(-3447)  # Pa, equal to -0.5 psi

    m.fs.air_compressor_s2.outlet.pressure.fix(206843)  # Pa, equal to 30 psia
    m.fs.air_compressor_s2.efficiency_isentropic.fix(0.84)

    m.fs.intercooler_s2.outlet.temperature.fix(310.93)  # K, equal to 100 F
    m.fs.intercooler_s2.deltaP.fix(-3447)  # Pa, equal to -0.5 psi

    # steam to reformer
    m.fs.reformer_mix.steam_inlet.flow_mol[0] == 464.77  # mol/s
    m.fs.reformer_mix.steam_inlet.temperature.fix(422)  # K
    m.fs.reformer_mix.steam_inlet.pressure.fix(206843)  # Pa, equal to 30 psia
    m.fs.reformer_mix.steam_inlet.mole_frac_comp.fix(0)
    m.fs.reformer_mix.steam_inlet.mole_frac_comp[0, 'H2O'].fix(1)

    # reformer outlet pressure
    m.fs.reformer.outlet.pressure.fix(137895)  # Pa, equal to 20 Psi
    m.fs.reformer.outlet.temperature.fix(1060.93)  # K, equal to 1450 F


def scale_flowsheet(m):
    # set NG_props default scaling
    m.fs.NG_props.set_default_scaling("flow_mol", 1e-3)
    m.fs.NG_props.set_default_scaling("flow_mol_phase", 1e-3)
    m.fs.NG_props.set_default_scaling("temperature", 1e-2)
    m.fs.NG_props.set_default_scaling("pressure", 1e-5)

    m.fs.NG_props.set_default_scaling("mole_frac_comp", 1e1)
    m.fs.NG_props.set_default_scaling("mole_frac_comp", 1e2, index="C2H6")
    m.fs.NG_props.set_default_scaling("mole_frac_comp", 1e2, index="C3H8")
    m.fs.NG_props.set_default_scaling("mole_frac_comp", 1e2, index="C4H10")

    m.fs.NG_props.set_default_scaling("mole_frac_phase_comp", 1e1)
    m.fs.NG_props.set_default_scaling("mole_frac_phase_comp", 1e2,
                                      index=('Vap', 'C2H6'))
    m.fs.NG_props.set_default_scaling("mole_frac_phase_comp", 1e2,
                                      index=('Vap', 'C3H8'))
    m.fs.NG_props.set_default_scaling("mole_frac_phase_comp", 1e2,
                                      index=('Vap', 'C4H10'))

    m.fs.NG_props.set_default_scaling("enth_mol_phase", 1e-3)
    m.fs.NG_props.set_default_scaling("entr_mol_phase", 1e-1)

    # set syn_props default scaling
    m.fs.syn_props.set_default_scaling("flow_mol", 1e-3)
    m.fs.syn_props.set_default_scaling("flow_mol_phase", 1e-3)
    m.fs.syn_props.set_default_scaling("temperature", 1e-2)
    m.fs.syn_props.set_default_scaling("pressure", 1e-5)
    m.fs.syn_props.set_default_scaling("mole_frac_comp", 1e1)
    m.fs.syn_props.set_default_scaling("mole_frac_phase_comp", 1e1)
    m.fs.syn_props.set_default_scaling("enth_mol_phase", 1e-3)
    m.fs.syn_props.set_default_scaling("entr_mol_phase", 1e-1)

    # set air_props default scaling
    m.fs.air_props.set_default_scaling("flow_mol", 1e-3)
    m.fs.air_props.set_default_scaling("flow_mol_phase", 1e-3)
    m.fs.air_props.set_default_scaling("temperature", 1e-2)
    m.fs.air_props.set_default_scaling("pressure", 1e-5)
    m.fs.air_props.set_default_scaling("mole_frac_comp", 1e1)
    m.fs.air_props.set_default_scaling("mole_frac_phase_comp", 1e1)
    m.fs.air_props.set_default_scaling("enth_mol_phase", 1e-3)
    m.fs.air_props.set_default_scaling("entr_mol_phase", 1e-3)

    iscale.set_scaling_factor(m.fs.prereformer.lagrange_mult, 1e-4)
    iscale.set_scaling_factor(m.fs.anode.lagrange_mult, 1e-4)

    iscale.calculate_scaling_factors(m)
    # apply scaling factors
    m.fs.NG_props.set_default_scaling("flow_mol", 1e-3)
    m.fs.NG_props.set_default_scaling("flow_mol_phase", 1e-3)
    m.fs.NG_props.set_default_scaling("temperature", 1e-2)
    m.fs.NG_props.set_default_scaling("pressure", 1e-5)

    m.fs.NG_props.set_default_scaling("mole_frac_comp", 1e1)
    m.fs.NG_props.set_default_scaling("mole_frac_comp", 1e2, index="C2H6")
    m.fs.NG_props.set_default_scaling("mole_frac_comp", 1e2, index="C3H8")
    m.fs.NG_props.set_default_scaling("mole_frac_comp", 1e2, index="C4H10")

    m.fs.NG_props.set_default_scaling("mole_frac_phase_comp", 1e1)
    m.fs.NG_props.set_default_scaling("mole_frac_phase_comp", 1e2,
                                      index=('Vap', 'C2H6'))
    m.fs.NG_props.set_default_scaling("mole_frac_phase_comp", 1e2,
                                      index=('Vap', 'C3H8'))
    m.fs.NG_props.set_default_scaling("mole_frac_phase_comp", 1e2,
                                      index=('Vap', 'C4H10'))

    m.fs.NG_props.set_default_scaling("enth_mol_phase", 1e-3)
    m.fs.NG_props.set_default_scaling("entr_mol_phase", 1e-1)

    iscale.set_scaling_factor(m.fs.reformer.lagrange_mult, 1e-4)

    iscale.calculate_scaling_factors(m)


def initialize_power_island(m):
    # cathode side
    m.fs.air_blower.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.cathode_hx.tube_inlet, m.fs.air_blower.outlet)

    # fix cathode inlet to initial guess
    m.fs.cathode.inlet.flow_mol[0] = 34174
    m.fs.cathode.inlet.temperature[0] = 892
    m.fs.cathode.inlet.pressure[0] = 105490
    m.fs.cathode.inlet.mole_frac_comp[0, 'H2O'] = 0.0109
    m.fs.cathode.inlet.mole_frac_comp[0, 'CO2'] = 0.0003
    m.fs.cathode.inlet.mole_frac_comp[0, 'N2'] = 0.8099
    m.fs.cathode.inlet.mole_frac_comp[0, 'O2'] = 0.1690
    m.fs.cathode.inlet.mole_frac_comp[0, 'Ar'] = 0.0099

    m.fs.cathode.initialize(outlvl=logging.INFO)

    m.fs.cathode.inlet.unfix()

    # cathode translator block
    copy_port_values(
        m.fs.cathode_translator.inlet, m.fs.cathode.ion_outlet)

    m.fs.cathode_translator.initialize()

    # rest of cathode side
    copy_port_values(
        m.fs.cathode_heat.inlet, m.fs.cathode.air_outlet)

    m.fs.cathode_heat.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.cathode_recycle.inlet, m.fs.cathode_heat.outlet)

    m.fs.cathode_recycle.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.cathode_hx.shell_inlet, m.fs.cathode_recycle.exhaust)

    m.fs.cathode_hx.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.cathode_blower.inlet, m.fs.cathode_recycle.recycle)

    m.fs.cathode_blower.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.cathode_mix.recycle, m.fs.cathode_blower.outlet)

    copy_port_values(
        m.fs.cathode_mix.feed, m.fs.cathode_hx.tube_outlet)

    m.fs.cathode_mix.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.cathode.inlet, m.fs.cathode_mix.outlet)

    # anode side
    # anode inlet is used as tear stream
    m.fs.anode.inlet.flow_mol[0] = 9702
    m.fs.anode.inlet.temperature[0] = 844
    m.fs.anode.inlet.pressure[0] = 137612
    m.fs.anode.inlet.mole_frac_comp[0, 'CH4'] = 0.058
    m.fs.anode.inlet.mole_frac_comp[0, 'CO'] = 0.044
    m.fs.anode.inlet.mole_frac_comp[0, 'CO2'] = 0.123
    m.fs.anode.inlet.mole_frac_comp[0, 'H2'] = 0.223
    m.fs.anode.inlet.mole_frac_comp[0, 'H2O'] = 0.183
    m.fs.anode.inlet.mole_frac_comp[0, 'N2'] = 0.195
    m.fs.anode.inlet.mole_frac_comp[0, 'O2'] = 0.172
    m.fs.anode.inlet.mole_frac_comp[0, 'Ar'] = 0.002

    m.fs.anode.lagrange_mult[0, "C"] = 51092
    m.fs.anode.lagrange_mult[0, "H"] = 78296
    m.fs.anode.lagrange_mult[0, "O"] = 291784

    m.fs.anode.outlet.mole_frac_comp[0, "O2"] = 0

    m.fs.anode.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.anode_recycle.inlet, m.fs.anode.outlet)

    m.fs.anode_recycle.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.anode_blower.inlet, m.fs.anode_recycle.recycle)

    m.fs.anode_blower.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.recycle_translator.inlet, m.fs.anode_blower.outlet)

    m.fs.recycle_translator.initialize()

    copy_port_values(
        m.fs.anode_mix.recycle, m.fs.recycle_translator.outlet)

    m.fs.anode_mix.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.anode_hx.tube_inlet, m.fs.anode_mix.outlet)

    copy_port_values(
        m.fs.anode_hx.shell_inlet, m.fs.anode_recycle.exhaust)

    m.fs.anode_hx.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.prereformer.inlet, m.fs.anode_hx.tube_outlet)

    copy_port_values(
        m.fs.prereformer.outlet, m.fs.prereformer.inlet)

    m.fs.prereformer.gibbs_scaling = 1e-4

    m.fs.prereformer.lagrange_mult[0, "C"] = 9707
    m.fs.prereformer.lagrange_mult[0, "H"] = 62744
    m.fs.prereformer.lagrange_mult[0, "O"] = 293569

    m.fs.prereformer.outlet.mole_frac_comp[0, "O2"] = 0
    m.fs.prereformer.outlet.mole_frac_comp[0, "Ar"] = 0.003
    m.fs.prereformer.outlet.mole_frac_comp[0, "C2H6"] = 0
    m.fs.prereformer.outlet.mole_frac_comp[0, "C3H8"] = 0
    m.fs.prereformer.outlet.mole_frac_comp[0, "C4H10"] = 0

    m.fs.prereformer.initialize(outlvl=logging.INFO)

    copy_port_values(
        m.fs.anode_translator.inlet, m.fs.prereformer.outlet)

    m.fs.anode_translator.initialize()

    copy_port_values(
        m.fs.fuel_cell_mix.fuel_inlet, m.fs.anode_translator.outlet)

    copy_port_values(
        m.fs.fuel_cell_mix.ion_inlet, m.fs.cathode_translator.outlet)

    m.fs.fuel_cell_mix.initialize(outlvl=logging.INFO)

    ##############################
    # Combustor and HRSG section #
    ##############################

    # cathode side
    copy_port_values(m.fs.cathode_exhaust_split.inlet,
                     m.fs.cathode_hx.shell_outlet)

    m.fs.cathode_exhaust_split.initialize(outlvl=logging.INFO)

    copy_port_values(m.fs.cathode_expander.inlet,
                     m.fs.cathode_exhaust_split.exhaust_outlet)

    m.fs.cathode_expander.initialize(outlvl=logging.INFO)

    copy_port_values(m.fs.cathode_HRSG.inlet,
                     m.fs.cathode_expander.outlet)

    m.fs.cathode_HRSG.initialize(outlvl=logging.INFO)

    copy_port_values(m.fs.cathode_exhaust_translator.inlet,
                     m.fs.cathode_exhaust_split.combustor_outlet)

    m.fs.cathode_exhaust_translator.initialize()

    # anode side
    copy_port_values(m.fs.combustor_mix.cathode_inlet,
                     m.fs.cathode_exhaust_translator.outlet)

    copy_port_values(m.fs.combustor_mix.anode_inlet,
                     m.fs.anode_hx.shell_outlet)

    m.fs.combustor_mix.initialize(outlvl=logging.INFO)

    copy_port_values(m.fs.combustor.inlet,
                     m.fs.combustor_mix.outlet)

    m.fs.combustor.initialize()

    copy_port_values(m.fs.combustor_expander.inlet,
                     m.fs.combustor.outlet)

    m.fs.combustor_expander.initialize()

    copy_port_values(m.fs.anode_HRSG.inlet,
                     m.fs.combustor_expander.outlet)

    m.fs.anode_HRSG.initialize()


def initialize_reformer(m):
    m.fs.reformer.inlet.flow_mol[0] = 2262  # mol/s
    m.fs.reformer.inlet.temperature[0] = 470  # K
    m.fs.reformer.inlet.pressure[0] = 203395  # Pa
    m.fs.reformer.inlet.mole_frac_comp[0, 'CH4'] = 0.191
    m.fs.reformer.inlet.mole_frac_comp[0, 'C2H6'] = 0.006
    m.fs.reformer.inlet.mole_frac_comp[0, 'C3H8'] = 0.002
    m.fs.reformer.inlet.mole_frac_comp[0, 'C4H10'] = 0.001
    m.fs.reformer.inlet.mole_frac_comp[0, 'H2'] = 0
    m.fs.reformer.inlet.mole_frac_comp[0, 'CO'] = 0
    m.fs.reformer.inlet.mole_frac_comp[0, 'CO2'] = 0.002
    m.fs.reformer.inlet.mole_frac_comp[0, 'H2O'] = 0.212
    m.fs.reformer.inlet.mole_frac_comp[0, 'N2'] = 0.458
    m.fs.reformer.inlet.mole_frac_comp[0, 'O2'] = 0.122
    m.fs.reformer.inlet.mole_frac_comp[0, 'Ar'] = 0.006

    m.fs.reformer.lagrange_mult[0, 'C'] = 39230
    m.fs.reformer.lagrange_mult[0, 'H'] = 81252
    m.fs.reformer.lagrange_mult[0, 'O'] = 315049

    m.fs.reformer.outlet.mole_frac_comp[0, 'O2'] = 0
    m.fs.reformer.outlet.mole_frac_comp[0, 'Ar'] = 0.004
    m.fs.reformer.outlet.mole_frac_comp[0, 'CH4'] = 0.0005
    m.fs.reformer.outlet.mole_frac_comp[0, 'C2H6'] = 0
    m.fs.reformer.outlet.mole_frac_comp[0, 'C3H8'] = 0
    m.fs.reformer.outlet.mole_frac_comp[0, 'C4H10'] = 0

    m.fs.reformer.initialize()

    # reformer recuperator
    copy_port_values(
        m.fs.reformer_recuperator.shell_inlet, m.fs.reformer.outlet)

    m.fs.reformer_recuperator.initialize()

    # NG expander
    copy_port_values(
        m.fs.NG_expander.inlet, m.fs.reformer_recuperator.tube_outlet)

    m.fs.NG_expander.initialize()

    # reformer bypass
    copy_port_values(
        m.fs.reformer_bypass.inlet, m.fs.NG_expander.outlet)

    m.fs.reformer_bypass.initialize()

    # air compressor train
    m.fs.air_compressor_s1.initialize()

    copy_port_values(
        m.fs.intercooler_s1.inlet, m.fs.air_compressor_s1.outlet)

    m.fs.intercooler_s1.initialize()

    copy_port_values(
        m.fs.air_compressor_s2.inlet, m.fs.intercooler_s1.outlet)

    m.fs.air_compressor_s2.initialize()

    copy_port_values(
        m.fs.intercooler_s2.inlet, m.fs.air_compressor_s2.outlet)

    m.fs.intercooler_s2.initialize()

    # reformer mixer
    copy_port_values(
        m.fs.reformer_mix.oxygen_inlet, m.fs.intercooler_s2.outlet)

    copy_port_values(
        m.fs.reformer_mix.gas_inlet, m.fs.reformer_bypass.reformer_outlet)

    m.fs.reformer_mix.initialize()

    # bypass rejoin
    copy_port_values(
        m.fs.bypass_rejoin.syngas_inlet,
        m.fs.reformer_recuperator.shell_outlet)

    copy_port_values(
        m.fs.bypass_rejoin.bypass_inlet, m.fs.reformer_bypass.bypass_outlet)

    m.fs.bypass_rejoin.initialize()


def connect_reformer_to_power_island(m):
    m.fs.anode_mix.feed.unfix()
    m.fs.SYN_CONNECT = Arc(
        source=m.fs.bypass_rejoin.outlet, destination=m.fs.anode_mix.feed)
    pyo.TransformationFactory("network.expand_arcs").apply_to(m.fs)


def SOFC_ROM_setup(m):
    # create the ROM
    build_SOFC_ROM(m.fs)

    # build constraints connecting flowsheet to ROM input vars

    @m.fs.Constraint()
    def ROM_fuel_inlet_temperature(fs):
        return(fs.SOFC.fuel_temperature ==
               fs.anode_mix.feed.temperature[0]/pyunits.K - 273)

    @m.fs.Constraint()
    def ROM_air_inlet_temperature(fs):
        return(fs.SOFC.air_temperature ==
               fs.cathode.inlet.temperature[0]/pyunits.K - 273)

    @m.fs.Constraint()
    def ROM_air_recirculation(fs):
        return(fs.SOFC.air_recirculation ==
               fs.cathode_recycle.split_fraction[0, 'recycle'])

    @m.fs.Constraint()
    def ROM_OTC(fs):
        O_frac = (1 * fs.fuel_cell_mix.fuel_inlet.mole_frac_comp[0, 'CO']
                  + 2 * fs.fuel_cell_mix.fuel_inlet.mole_frac_comp[0, 'CO2']
                  + 1 * fs.fuel_cell_mix.fuel_inlet.mole_frac_comp[0, 'H2O'])

        C_frac = (1 * fs.fuel_cell_mix.fuel_inlet.mole_frac_comp[0, 'CO']
                  + 1 * fs.fuel_cell_mix.fuel_inlet.mole_frac_comp[0, 'CO2']
                  + 1 * fs.fuel_cell_mix.fuel_inlet.mole_frac_comp[0, 'CH4'])

        return fs.SOFC.OTC == O_frac/C_frac

    @m.fs.Constraint()
    def ROM_fuel_utilization(fs):
        full_O2_flow = (fs.anode_mix.feed.flow_mol[0] * (
            0.5 * fs.anode_mix.feed.mole_frac_comp[0, 'H2'] +
            0.5 * fs.anode_mix.feed.mole_frac_comp[0, 'CO'] +
            2.0 * fs.anode_mix.feed.mole_frac_comp[0, 'CH4'] +
            3.5 * fs.anode_mix.feed.mole_frac_comp[0, 'C2H6'] +
            5.0 * fs.anode_mix.feed.mole_frac_comp[0, 'C3H8'] +
            6.5 * fs.anode_mix.feed.mole_frac_comp[0, 'C4H10']))

        return (fs.SOFC.fuel_util*full_O2_flow ==
                fs.cathode.ion_outlet.flow_mol[0])

    @m.fs.Constraint()
    def ROM_air_utilization(fs):
        air_in = (fs.cathode_hx.tube_inlet.flow_mol[0] *
                  fs.cathode_hx.tube_inlet.mole_frac_comp[0, 'O2'])

        air_out = (fs.cathode_hx.shell_inlet.flow_mol[0] *
                   fs.cathode_hx.shell_inlet.mole_frac_comp[0, 'O2'])

        return fs.SOFC.air_util == 1 - air_out/air_in

    @m.fs.Constraint()
    def ROM_internal_reformation(fs):
        return (fs.SOFC.internal_reforming ==
                fs.reformer_bypass.split_fraction[0, 'bypass_outlet'])

    # we want to analyze this flowsheet in terms of ROM inputs, so fix ROM
    # inputs and unfix flowsheet vars that control inputs

    # current density
    m.fs.SOFC.current_density.fix(4000)

    # fuel temperature
    m.fs.reformer_recuperator.tube_outlet.temperature.unfix()
    m.fs.SOFC.fuel_temperature.fix(348.33)

    # internal reformation fraction
    m.fs.reformer_bypass.split_fraction[0, 'bypass_outlet'].unfix()
    m.fs.SOFC.internal_reforming.fix(0.6)

    # air temperature - constrained by max cell temperature
    m.fs.cathode_hx.area.unfix()
    m.fs.SOFC.max_cell_temperature.fix(750)

    # air recirculation fraction
    m.fs.cathode_recycle.split_fraction[0, 'recycle'].unfix()
    m.fs.SOFC.air_recirculation.fix(0.5)

    # oxygen to carbon ratio
    m.fs.anode_recycle.split_fraction[0, 'recycle'].unfix()
    m.fs.SOFC.OTC.fix(2.1)

    # fuel utilization
    m.fs.cathode.ion_outlet.flow_mol.unfix()
    m.fs.SOFC.fuel_util.fix(0.8)

    # air utilization - constrained by deltaT across cell
    m.fs.air_blower.inlet.flow_mol.unfix()
    m.fs.SOFC.deltaT_cell.fix(100)

    # pressure
    m.fs.SOFC.pressure.fix(1)

    # initialize ROM
    calculate_variable_from_constraint(m.fs.SOFC.air_temperature,
                                       m.fs.ROM_air_inlet_temperature)
    calculate_variable_from_constraint(m.fs.SOFC.air_util,
                                       m.fs.ROM_air_utilization)
    initialize_SOFC_ROM(m.fs.SOFC)

    # add constraints for power calculations
    m.fs.F = pyo.Param(initialize=96487, units=pyunits.C/pyunits.mol)
    m.fs.stack_current = pyo.Var(initialize=650, units=pyunits.MA)
    m.fs.stack_power = pyo.Var(initialize=600, units=pyunits.MW)

    @m.fs.Constraint()
    def stack_current_constraint(fs):
        return (fs.stack_current ==
                pyunits.convert(4*fs.F*m.fs.cathode.ion_outlet.flow_mol[0],
                                pyunits.MA))

    @m.fs.Constraint()
    def stack_power_constraint(fs):
        return fs.stack_power == fs.stack_current*fs.SOFC.stack_voltage

    # initialize power calculations
    calculate_variable_from_constraint(m.fs.stack_current,
                                       m.fs.stack_current_constraint)
    calculate_variable_from_constraint(m.fs.stack_power,
                                       m.fs.stack_power_constraint)


def add_SOFC_energy_balance(m):
    @m.fs.Constraint()
    def ROM_anode_outlet_temperature(fs):
        return (fs.SOFC.anode_outlet_temperature ==
                fs.anode.outlet.temperature[0]/pyunits.K - 273)

    m.fs.anode.outlet.temperature.unfix()

    @m.fs.Constraint()
    def SOFC_energy_balance(fs):
        return (-1*pyunits.convert(fs.anode.heat_duty[0], pyunits.MW) ==
                fs.stack_power +
                pyunits.convert(fs.cathode_heat.heat_duty[0], pyunits.MW))

    m.fs.cathode_heat.outlet.temperature.unfix()


def add_result_constraints(m):
    # total heat supplied to HRSG
    m.fs.HRSG_heat_duty = pyo.Var(initialize=300, units=pyunits.MW)

    @m.fs.Constraint()
    def HRSG_heat_duty_constraint(fs):
        return (-1*fs.HRSG_heat_duty ==
                pyunits.convert(fs.anode_HRSG.heat_duty[0], pyunits.MW) +
                pyunits.convert(fs.cathode_HRSG.heat_duty[0], pyunits.MW))

    # heat required for generating reformer steam
    m.fs.reformer_steam_heat = pyo.Var(initialize=25, units=pyunits.MW)

    # only valid when steam is heated to 422 K (149 C)
    @m.fs.Constraint()
    def reformer_steam_heat_constraint(fs):
        return (fs.reformer_steam_heat == 0.050735*pyunits.MJ/pyunits.mol *
                fs.reformer_mix.steam_inlet.flow_mol[0])

    # HRSG heat applied toward steam cycle
    m.fs.steam_cycle_heat = pyo.Var(initialize=300, units=pyunits.MW)

    @m.fs.Constraint()
    def steam_cycle_heat_constraint(fs):
        return (fs.steam_cycle_heat == fs.HRSG_heat_duty -
                fs.reformer_steam_heat -
                pyunits.convert(fs.reformer.heat_duty[0], pyunits.MW))

    # power generated by steam cycle
    m.fs.steam_cycle_power = pyo.Var(initialize=100, units=pyunits.MW)
    m.fs.steam_cycle_efficiency = pyo.Param(initialize=0.38, mutable=True)

    @m.fs.Constraint()
    def steam_cycle_power_constraint(fs):
        return (fs.steam_cycle_power ==
                fs.steam_cycle_heat * fs.steam_cycle_efficiency)

    # stack AC power
    m.fs.stack_power_AC = pyo.Var(initialize=550, units=pyunits.MW)
    m.fs.inverter_efficiency = pyo.Param(initialize=0.97, mutable=True)

    @m.fs.Constraint()
    def stack_AC_power_constraint(fs):
        return fs.stack_power_AC == fs.stack_power * fs.inverter_efficiency

    # gross plant power
    m.fs.gross_power = pyo.Var(initialize=670, units=pyunits.MW)

    @m.fs.Constraint()
    def gross_power_constraint(fs):
        return (fs.gross_power == fs.stack_power_AC +
                fs.steam_cycle_power +
                -1*pyunits.convert(fs.NG_expander.work_mechanical[0],
                                   pyunits.MW))

    # auxiliary load of the plant
    m.fs.auxiliary_load = pyo.Var(initialize=10, units=pyunits.MW)

    @m.fs.Constraint()
    def auxiliary_load_constraint(fs):
        return (fs.auxiliary_load == pyunits.convert(
                   (fs.air_blower.work_mechanical[0] +
                    fs.cathode_blower.work_mechanical[0] +
                    fs.anode_blower.work_mechanical[0] +
                    fs.air_compressor_s1.work_mechanical[0] +
                    fs.air_compressor_s2.work_mechanical[0]), pyunits.MW))

    # net plant power
    m.fs.net_power = pyo.Var(initialize=660, units=pyunits.MW)

    @m.fs.Constraint()
    def net_power_constraint(fs):
        return fs.net_power == fs.gross_power - fs.auxiliary_load

    # HHV efficiency
    m.fs.HHV_efficiency = pyo.Var(initialize=0.6)

    @m.fs.Constraint()
    def efficiency_rule(fs):
        NG_HHV = 908839.23*pyunits.J/pyunits.mol
        return (fs.HHV_efficiency == fs.net_power /
                pyunits.convert(
                    (NG_HHV * fs.reformer_recuperator.tube_inlet.flow_mol[0]),
                    pyunits.MW))

    # CO2 emissions in g/kWh
    m.fs.CO2_emissions = pyo.Var(initialize=300,
                                 units=pyunits.g/pyunits.kWh)

    @m.fs.Constraint()
    def CO2_emission_constraint(fs):
        mass_flow = (fs.anode_HRSG.outlet.flow_mol[0] *
                     fs.anode_HRSG.outlet.mole_frac_comp[0, 'CO2'] *
                     44.01*pyunits.g/pyunits.mol)
        return fs.CO2_emissions == pyunits.convert((mass_flow/fs.net_power),
                                                   (pyunits.g/pyunits.kWh))


def make_stream_dict(m):
    """Adds _streams to m, which contains a dictionary of streams for display

    Args:
        m (ConcreteModel): A Pyomo model from create_model()

    Returns:
        None
    """

    m._streams = OrderedDict(
        [
            ("FUEL_IN", m.fs.reformer_recuperator.tube_inlet),
            ("HOT_FUEL", m.fs.reformer_recuperator.tube_outlet),
            ("REF_IN", m.fs.reformer_mix.gas_inlet),
            ("AIR_IN", m.fs.reformer_mix.oxygen_inlet),
            ("STEAM_IN", m.fs.reformer_mix.steam_inlet),
            ("REF_OUT", m.fs.reformer.outlet),
            ("SYN_IN", m.fs.anode_mix.feed),
            ("ANO_HX_CI", m.fs.ANODE_MIXER),
            ("ANODE_IN", m.fs.fuel_cell_mix.fuel_inlet),
            ("ANODE_OUT", m.fs.anode.outlet),
            ("ANODE_REC", m.fs.ANODE_BLOWER),
            ("ANO_HX_HI", m.fs.ANODE_RECYCLE_HX),
            ("ANO_HX_HO", m.fs.anode_hx.shell_outlet),
            ("AIR", m.fs.air_blower.inlet),
            ("CATH_HX_CI", m.fs.AIR_BLOWER),
            ("CATH_HX_CO", m.fs.CATHODE_HX_COLD),
            ("CATH_IN", m.fs.CATHODE_MIXER),
            ("CATH_OUT", m.fs.CATHODE_HEAT),
            ("CATH_REC", m.fs.CATHODE_BLOWER),
            ("CATH_HX_HI", m.fs.CATHODE_RECYCLE_HX),
            ("CATH_HX_HO", m.fs.cathode_hx.shell_outlet),
            ("CATH_HRSG_IN", m.fs.cathode_HRSG.inlet),
            ("CATH_EXH", m.fs.cathode_HRSG.outlet),
            ("COMB_AIR", m.fs.combustor_mix.cathode_inlet),
            ("COMB_OUT", m.fs.combustor.outlet),
            ("ANOD_EXH", m.fs.anode_HRSG.outlet)])


def pfd_result(outfile, m, df):
    # a function for string formatting
    def fstr(value, decimals, unit=''):
        if decimals == 0:
            rounded_value = int(value)
        else:
            rounded_value = round(value, decimals)
        return "{:,}".format(rounded_value) + unit
    # set up tags
    tags = {}
    for i in df.index:
        tags[i + "_F"] = fstr(df.loc[i, "Total Molar Flowrate"], 0, ' mol/s')
        tags[i + "_T"] = fstr(df.loc[i, "Temperature"], 0, ' K')
        tags[i + "_P"] = fstr(df.loc[i, "Pressure"]/1000, 0, ' kPa')
    tags["current_density"] = fstr(pyo.value(m.fs.SOFC.current_density), 0)
    tags["fuel_temp"] = fstr(pyo.value(m.fs.SOFC.fuel_temperature), 1)
    tags["int_ref"] = fstr(pyo.value(m.fs.SOFC.internal_reforming), 2)
    tags["air_temp"] = fstr(pyo.value(m.fs.SOFC.air_temperature), 1)
    tags["air_recirc"] = fstr(pyo.value(m.fs.SOFC.air_recirculation), 2)
    tags["OTC"] = fstr(pyo.value(m.fs.SOFC.OTC), 3)
    tags["fuel_util"] = fstr(pyo.value(m.fs.SOFC.fuel_util), 3)
    tags["air_util"] = fstr(pyo.value(m.fs.SOFC.air_util), 3)
    tags["voltage"] = fstr(pyo.value(m.fs.SOFC.stack_voltage), 4)
    tags["power"] = fstr(pyo.value(m.fs.stack_power), 1)
    tags["recuperator_duty"] = fstr(
        (pyo.value(m.fs.reformer_recuperator.heat_duty[0])/1e6), 1)
    tags["anode_hx_duty"] = fstr(
        (pyo.value(m.fs.anode_hx.heat_duty[0])/1e6), 2)
    tags["cathode_hx_duty"] = fstr(
        (pyo.value(m.fs.cathode_hx.heat_duty[0])/1e6), 2)
    tags["anode_duty"] = fstr((pyo.value(m.fs.anode.heat_duty[0])/1e6), 2)
    tags["cathode_duty"] = fstr(
        (pyo.value(m.fs.cathode_heat.heat_duty[0])/1e6), 2)
    tags["stack_power_AC"] = fstr(pyo.value(m.fs.stack_power_AC), 1)
    tags["HRSG_power"] = fstr(pyo.value(m.fs.steam_cycle_power), 1)
    tags["expander_power"] = fstr(-1*pyo.value(
        m.fs.NG_expander.work_mechanical[0])/1e6, 1)
    tags["aux_load"] = fstr(pyo.value(m.fs.auxiliary_load), 1)
    tags["net_power"] = fstr(pyo.value(m.fs.net_power), 1)
    tags["thermal_input"] = fstr(pyo.value(
        m.fs.net_power/m.fs.HHV_efficiency), 1)
    tags["efficiency"] = fstr(pyo.value(m.fs.HHV_efficiency)*100, 2)
    tags["emissions"] = fstr(pyo.value(m.fs.CO2_emissions), 1)

    original_svg_file = os.path.join(this_file_dir(),
                                     "NGFC_results_template.svg")
    with open(original_svg_file, "r") as f:
        svg_tag(tags, f, outfile=outfile)


def main():
    # create model and flowsheet
    m = pyo.ConcreteModel(name='NGFC without carbon capture')
    m.fs = FlowsheetBlock(default={"dynamic": False})

    # solver and options
    solver = pyo.SolverFactory("ipopt")
    solver.options = {'bound_push': 1e-16}

    if os.path.exists('NGFC_flowsheet_init.json.gz'):
        build_power_island(m)
        build_reformer(m)
        connect_reformer_to_power_island(m)
        SOFC_ROM_setup(m)
        add_SOFC_energy_balance(m)
        add_result_constraints(m)
        scale_flowsheet(m)
        ms.from_json(m, fname='NGFC_flowsheet_init.json.gz')

    else:
        build_power_island(m)
        build_reformer(m)
        set_power_island_inputs(m)
        set_reformer_inputs(m)
        scale_flowsheet(m)
        initialize_power_island(m)
        initialize_reformer(m)
        connect_reformer_to_power_island(m)
        SOFC_ROM_setup(m)
        add_SOFC_energy_balance(m)
        add_result_constraints(m)
        solver.solve(m, tee=True)
        ms.to_json(m, fname='NGFC_flowsheet_init.json.gz')

    # uncomment to report results
    # make_stream_dict(m)
    # df = create_stream_table_dataframe(streams=m._streams, orient="index")
    # pfd_result("NGFC_results.svg", m, df)

    return m



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for eNRTL methods

References:

[1] Local Composition Model for Excess Gibbs Energy of Electrolyte Systems, Pt 1.
Chen, C.-C., Britt, H.I., Boston, J.F., Evans, L.B.,
AIChE Journal, 1982, Vol. 28(4), pgs. 588-596

[2] New Data on Activity Coefficients of Potassium, Nitrate, and Chloride Ions
in Aqueous Solutions of KNO3 and KCl by Ion Selective Electrodes
Dash, D., Kumar, S., Mallika, C., Kamachi Mudali, U.,
ISRN Chemical Engineering, 2012, doi:10.5402/2012/730154

Author: Andrew Lee
"""
import pytest
from math import log

from pyomo.environ import (ConcreteModel,
                           units as pyunits,
                           value)

from idaes.core import (AqueousPhase,
                        Solvent,
                        Apparent,
                        Anion,
                        Cation)
from idaes.generic_models.properties.core.eos.enrtl import ENRTL
from idaes.generic_models.properties.core.eos.enrtl_reference_states import \
    Unsymmetric
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock, StateIndex)
from idaes.generic_models.properties.core.state_definitions import FTPx
from idaes.generic_models.properties.core.pure.electrolyte import \
    relative_permittivity_constant


def dummy_method(b, *args, **kwargs):
    return 1000/18e-3*pyunits.mol/pyunits.m**3


configuration = {
    "components": {
        "H2O": {"type": Solvent,
                "dens_mol_liq_comp": dummy_method,
                "relative_permittivity_liq_comp":
                    relative_permittivity_constant,
                "parameter_data": {
                    "mw": (18E-3, pyunits.kg/pyunits.mol),
                    "relative_permittivity_liq_comp": 78.54}},
        "KCl": {"type": Apparent,
                "dissociation_species": {"K+": 1, "Cl-": 1}},
        "K+": {"type": Cation,
               "charge": +1},
        "Cl-": {"type": Anion,
                "charge": -1}},
    "phases": {
        "Liq": {"type": AqueousPhase,
                "equation_of_state": ENRTL,
                "equation_of_state_options": {
                    "reference_state": Unsymmetric}}},
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "state_definition": FTPx,
    "state_components": StateIndex.true,
    "pressure_ref": 1e5,
    "temperature_ref": 300,
    "parameter_data": {
        "Liq_tau": {
            ("H2O", "K+, Cl-"): 8.064,  # Table 1, [1]
            ("K+, Cl-", "H2O"): -4.107}}}  # Table 1, [1]


class TestStateBlockUnsymmetric(object):
    @pytest.fixture(scope="class")
    def model(self):
        config = dict(configuration)
        eos_opt = config["phases"]["Liq"]["equation_of_state_options"] = {}
        eos_opt["reference_state"] = Unsymmetric

        m = ConcreteModel()
        m.params = GenericParameterBlock(default=config)

        m.state = m.params.build_state_block([1])

        # Need to set a value of T for checking expressions later
        m.state[1].temperature.set_value(298)

        return m

    @pytest.mark.unit
    def test_parameters(self, model):
        assert model.params.Liq.tau["H2O", "K+, Cl-"].value == 8.064
        assert model.params.Liq.tau["K+, Cl-", "H2O"].value == -4.107

    @pytest.mark.unit
    def test_log_gamma(self, model):
        # start with pure water
        # Using 0 results in division by zero errors
        for k in model.state[1].mole_frac_phase_comp:
            model.state[1].mole_frac_phase_comp[k].set_value(1e-12)

        # Data from [2] - Form {molality: gamma_KCl}
        data = {0.001: 0.965,
                0.002: 0.951,
                0.005: 0.927,
                0.01: 0.901,
                0.0201: 0.869,
                0.05: 0.816,
                0.0982: 0.768,
                0.1983: 0.717,
                0.3049: 0.687,
                0.3999: 0.665,
                0.5001: 0.649,
                0.6032: 0.636,
                0.7003: 0.626,
                0.7996: 0.617,
                0.8968: 0.61,
                0.9926: 0.604,
                1.1899: 0.594,
                1.3844: 0.586,
                1.5817: 0.58,
                1.7923: 0.576,
                1.9895: 0.573,
                2.5039: 0.568,
                2.9837: 0.568,
                3.4982: 0.571,
                3.994: 0.576,
                4.4897: 0.584,
                4.7909: 0.589,
                4.9908: 0.593
                }

        for x, g in data.items():
            w = 1000/18

            model.state[1].mole_frac_phase_comp["Liq", "H2O"].set_value(
                w/(w+2*x))
            model.state[1].mole_frac_phase_comp["Liq", "K+"].set_value(
                x/(w+2*x))
            model.state[1].mole_frac_phase_comp["Liq", "Cl-"].set_value(
                x/(w+2*x))

            conv = log(1+18*2*x/1000)  # Convert mole frac to molal basis
            assert pytest.approx(log(g), rel=3e-2, abs=6e-3) == value(
                model.state[1].Liq_log_gamma["K+"] - conv)
            assert pytest.approx(log(g), rel=3e-2, abs=6e-3) == value(
                model.state[1].Liq_log_gamma["Cl-"] - conv)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Tests for turbine stage model.

Author: John Eslick
"""
import pytest

from pyomo.environ import ConcreteModel

from idaes.core import FlowsheetBlock
from idaes.power_generation.unit_models.helm import HelmTurbineStage
from idaes.generic_models.properties import iapws95
from idaes.core.util.model_statistics import (
        degrees_of_freedom,
        activated_equalities_generator)
from idaes.core.util import get_solver

# Set up solver
solver = get_solver()


@pytest.fixture()
def build_turbine():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.turb = HelmTurbineStage(default={"property_package": m.fs.properties})
    return m


@pytest.fixture()
def build_turbine_dyn():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": True})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.turb = HelmTurbineStage(default={
        "dynamic": False,
        "property_package": m.fs.properties})
    return m


@pytest.mark.unit
def test_basic_build(build_turbine):
    """Make a turbine model and make sure it doesn't throw exception"""
    m = build_turbine


@pytest.mark.component
def test_initialize(build_turbine):
    """Initialize a turbine model"""
    m = build_turbine
    # set inlet
    m.fs.turb.inlet.enth_mol[0].value = 70000
    m.fs.turb.inlet.flow_mol[0].value = 15000
    m.fs.turb.inlet.pressure[0].value = 8e6
    m.fs.turb.efficiency_isentropic.fix(0.8)
    m.fs.turb.ratioP.fix(0.7)
    m.fs.turb.initialize(outlvl=4)

    eq_cons = activated_equalities_generator(m)
    for c in eq_cons:
        assert(abs(c.body() - c.lower) < 1e-4)
    assert(degrees_of_freedom(m)==3) #inlet was't fixed and still shouldn't be


@pytest.mark.unit
def test_report(build_turbine):
    m = build_turbine
    m.fs.turb.report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################

# stdlib
from collections import namedtuple
from pathlib import Path
import sys
import time
from typing import Optional, Union, Tuple
import webbrowser

# package
from idaes import logger
from .model_server import FlowsheetServer
from . import persist, errors

# Logging
_log = logger.getLogger(__name__)

# Module globals
web_server = None

#: Maximum number of saved versions of the same `save` file.
#: Set to zero if you want to allow any number.
MAX_SAVED_VERSIONS = 100


# Classes and functions

#: Return value for `visualize()` function. This namedtuple has three
#: attributes that can be accessed by position or name:
#:
#: - store = :class:`idaes.ui.fsvis.persist.DataStore` object (with a ``.filename`` attribute)
#: - port = Port number (integer) where web server is listening
#: - server = :class:`idaes.ui.fsvis.model_server.FlowsheetServer` object for the web server thread
#:
VisualizeResult = namedtuple("VisualizeResult", ["store", "port", "server"])


def visualize(
    flowsheet,
    name: str = "flowsheet",
    save: Optional[Union[Path, str, bool]] = None,
    save_dir: Optional[Path] = None,
    overwrite: bool = False,
    browser: bool = True,
    port: Optional[int] = None,
    log_level: int = logger.WARNING,
    quiet: bool = False,
    loop_forever: bool = False,
) -> VisualizeResult:
    """Visualize the flowsheet in a web application.

    The web application is started in a separate thread and this function returns immediately.

    Also open a browser window to display the visualization app. The URL is printed unless ``quiet`` is True.

    Args:
        flowsheet: IDAES flowsheet to visualize
        name: Name of flowsheet to display as the title of the visualization
        save: Where to save the current flowsheet layout and values. If this argument is not specified,
          "``name``.json" will be used (if this file already exists, a "-`<version>`" number will be added
          between the name and the extension). If the value given is the boolean 'False', then nothing
          will be saved. The boolean 'True' value is treated the same as unspecified.
        save_dir: If this argument is given, and ``save`` is not given or a relative path, then it will
           be used as the directory to save the default or given file. The current working directory is
           the default. If ``save`` is given and an absolute path, this argument is ignored.
        overwrite: If True, and the file given by ``save`` exists, overwrite instead of creating a new
          numbered file.
        browser: If true, open a browser
        port: Start listening on this port. If not given, find an open port.
        log_level: An IDAES logging level, which is a superset of the built-in :mod:`logging` module levels.
          See the :mod:`idaes.logger` module for details
        quiet: If True, suppress printing any messages to standard output (console)
        loop_forever: If True, don't return but instead loop until a Control-C is received. Useful when
           invoking this function at the end of a script.

    Returns:
        See :data:`VisualizeResult`

    Raises:
        :mod:`idaes.ui.fsvis.errors.VisualizerSaveError`: if the data storage at 'save_as' can't be opened
        :mod:`idaes.ui.fsvis.errors.VisualizerError`: Any other errors
        RuntimeError: If too many versions of the save file already exist. See :data:`MAX_SAVED_VERSIONS`.
    """
    global web_server

    # Initialize IDAES logging
    _init_logging(log_level)

    # Start the web server
    if web_server is None:
        web_server = FlowsheetServer(port=port)
        web_server.start()
        if not quiet:
            print("Started visualization server")
    else:
        _log.info(f"Using HTTP server on localhost, port {web_server.port}")

    # Set up save location
    use_default = False
    if save is None or save is True:
        save_path = _pick_default_save_location(name, save_dir)
        use_default = True
    elif save is False:
        save_path = None
    else:
        try:
            save_path = Path(save)
        except TypeError as err:
            raise errors.VisualizerSaveError(
                save, f"Cannot convert 'save' value to Path object: {err}"
            )
        if save_dir is not None and not save_path.is_absolute():
            save_path = save_dir / save_path
    # Create datastore for save location
    if save_path is None:
        datastore = persist.MemoryDataStore()
    else:
        # deal with duplicate names
        try:
            save_path = _handle_existing_save_path(
                name, save_path, max_versions=MAX_SAVED_VERSIONS, overwrite=overwrite
            )
        except errors.TooManySavedVersions as err:
            raise RuntimeError(f"In visualize(): {err}")
        datastore = persist.DataStore.create(save_path)
        if use_default:
            if not quiet:
                cwd = save_path.parent.absolute()
                print(
                    f"Saving flowsheet to default file '{save_path.name}' in current directory ({cwd})"
                )
        else:
            if not quiet:
                print(f"Saving flowsheet to {str(datastore)}")

    # Add our flowsheet to it
    try:
        new_name = web_server.add_flowsheet(name, flowsheet, datastore)
    except (errors.ProcessingError, errors.DatastoreError) as err:
        raise errors.VisualizerError(f"Cannot add flowsheet: {err}")

    if new_name != name:
        _log.warning(f"Flowsheet name changed: old='{name}' new='{new_name}'")
        if not quiet:
            print(f"Flowsheet name changed to '{new_name}'")
        name = new_name

    # Open a browser window for the UI
    url = f"http://localhost:{web_server.port}/app?id={name}"
    if browser:
        success = webbrowser.open(url)
        if success:
            _log.debug(f"Flowsheet opened in browser window")
        else:
            _log.warning(f"Could not open flowsheet URL '{url}' in browser")
            if not quiet:
                print("Error: Unable to open flowsheet in web browser.")

    if not quiet:
        print(f"Flowsheet visualization at: {url}")

    if loop_forever:
        _loop_forever(quiet)

    return VisualizeResult(store=datastore, port=web_server.port, server=web_server)


def _loop_forever(quiet):
    try:
        if not quiet:
            print("Type ^C to stop the program")
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        if not quiet:
            print("Program stopped")


def _pick_default_save_location(name, save_dir):
    """Pick a default save location."""
    if not save_dir:
        save_dir = Path(".")
    save_path = save_dir / f"{name}.json"
    return save_path


def _handle_existing_save_path(name, save_path, max_versions=10, overwrite=None):
    """Set up for overwrite/versioning for existing save paths."""
    save_dir = save_path.parent
    # Handle simple cases: overwrite, and no existing file
    if overwrite:
        if save_path.exists():
            _log.warning("Overwriting existing save file '{save_path}'")
            save_path.open("w")  # blank file
        return save_path
    elif not save_path.exists():
        return save_path
    # Find the next version that does not exist
    _log.info(f"Save file {save_path} exists. Creating new version")
    counter = 0
    if max_versions == 0:
        max_versions = sys.maxsize  # millions of years of file-creating fun
    while save_path.exists() and counter < max_versions:
        counter += 1
        save_file = f"{name}-{counter}.json"
        save_path = save_dir / save_file
    # Edge case: too many NAME-#.json files for this NAME
    if counter == max_versions:
        why = f"Found {max_versions} numbered files of form '{name}-<num>.json'. That's too many."
        _log.error(why)
        raise errors.TooManySavedVersions(why)
    # Return new (versioned) path
    _log.info(f"Created new version for save file: {save_path}")
    return save_path


def _init_logging(lvl):
    ui_logger = logger.getIdaesLogger("ui", level=lvl, tag="ui")
    ui_logger.setLevel(lvl)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Standard IDAES Feed block.
"""
# Import Pyomo libraries
from pyomo.environ import Reference
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        UnitModelBlockData,
                        useDefault)
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog


__author__ = "Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Feed")
class FeedData(UnitModelBlockData):
    """
    Standard Feed Block Class
    """
    CONFIG = ConfigBlock()
    CONFIG.declare("dynamic", ConfigValue(
        domain=In([False]),
        default=False,
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not,
**default** = False. Feed blocks are always steady-state."""))
    CONFIG.declare("has_holdup", ConfigValue(
        default=False,
        domain=In([False]),
        description="Holdup construction flag - must be False",
        doc="""Feed blocks do not contain holdup, thus this must be False."""))
    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))
    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}"""))

    def build(self):
        """
        Begin building model.

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(FeedData, self).build()

        # Add State Block
        self.properties = (
                    self.config.property_package.build_state_block(
                            self.flowsheet().time,
                            doc="Material properties in feed",
                            default={
                                "defined_state": True,
                                "has_phase_equilibrium": False,
                                **self.config.property_package_args}))

        # Add references to all state vars
        s_vars = self.properties[
                self.flowsheet().time.first()].define_state_vars()
        for s in s_vars:
            l_name = s_vars[s].local_name
            if s_vars[s].is_indexed():
                slicer = self.properties[:].component(l_name)[...]
            else:
                slicer = self.properties[:].component(l_name)

            r = Reference(slicer)
            setattr(self, s, r)

        # Add outlet port
        self.add_port(name="outlet", block=self.properties, doc="Outlet Port")

    def initialize(blk, state_args=None, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        '''
        This method calls the initialization method of the state block.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) to provide an initial state for
                           initialization (see documentation of the specific
                           property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        '''
        # ---------------------------------------------------------------------
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")

        if state_args is None:
            state_args = {}

        # Initialize state block
        blk.properties.initialize(outlvl=outlvl,
                                  optarg=optarg,
                                  solver=solver,
                                  state_args=state_args)

        init_log.info('Initialization Complete.')

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
                {"Outlet": self.outlet},
                time_point=time_point)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
"""
Visualization server back-end.

The main class is `FlowsheetServer`, which is instantiated from the `visualize()` function.
"""

# stdlib
import http.server
import json
from pathlib import Path
import re
import socket
import threading
from typing import Dict, Union
from urllib.parse import urlparse

# package
from idaes import logger
from ..flowsheet import FlowsheetDiff, FlowsheetSerializer
from . import persist, errors

_log = logger.getLogger(__name__)

# Directories
_this_dir = Path(__file__).parent.absolute()
_static_dir = _this_dir / "static"
_template_dir = _this_dir / "templates"


class FlowsheetServer(http.server.HTTPServer):
    """A simple HTTP server that runs in its own thread.

    This server is used for *all* models for a given process, so every request needs to contain
    the ID of the model that should be used in that transaction.

    The only methods that the visualization function needs to call are the constructor, `start()` to
     start running the server, and `add_flowsheet()`, to a add a new flowsheet.
    """

    def __init__(self, port=None):
        """Create HTTP server
        """
        self._port = port or find_free_port()
        _log.info(f"Starting HTTP server on localhost, port {self._port}")
        super().__init__(("127.0.0.1", self._port), FlowsheetServerHandler)
        self._dsm = persist.DataStoreManager()
        self._flowsheets = {}
        self._thr = None

    @property
    def port(self):
        return self._port

    def start(self):
        """Start the server, which will spawn a thread.
        """
        self._thr = threading.Thread(target=self._run)
        self._thr.setDaemon(True)
        self._thr.start()

    def add_flowsheet(self, id_, flowsheet, store: persist.DataStore) -> str:
        """Add a flowsheet, and also the method of saving it.

        Args:
            id_: Name of flowsheet
            flowsheet: Flowsheet object
            store: A DataStore for saving the flowsheet

        Returns:
            Name of flowsheet, modified as necessary to be URL friendly

        Raises:
            ProcessingError: if the flowsheet can't be serialized
            DatastoreError: If the flowsheet can't be saved
        """
        # replace all but 'unreserved' (RFC 3896) chars with a dash; remove duplicate dashes
        id_ = self.canonical_flowsheet_name(id_)
        self._flowsheets[id_] = flowsheet
        _log.debug(f"Flowsheet '{id_}' storage is {store}")
        self._dsm.add(id_, store)
        # First try to update, so as not to overwrite saved value
        try:
            self.update_flowsheet(id_)
        except errors.FlowsheetNotFoundInDatastore:
            _log.debug(f"No existing flowsheet found in {store}: saving new value")
            # If not found in datastore, save new value
            fs_dict = FlowsheetSerializer(flowsheet, id_).as_dict()
            store.save(fs_dict)
        else:
            _log.debug(f"Existing flowsheet found in {store}: saving merged value")
        return id_

    @staticmethod
    def canonical_flowsheet_name(name: str) -> str:
        """Create a canonical flowsheet name from the name provided by the user.

        Replace all but 'unreserved' (RFC 3896) chars plus '~' with a dash and remove duplicate dashes.
        The result will not have whitespace, slashes, punctuation, or any special characters.

        Args:
            name: User-provided name

        Returns:
            New name
        """
        return re.sub(r"-+", "-", re.sub(r"[^a-zA-Z0-9-._]", "-", name))

    # === Public methods called only by HTTP handler ===

    def save_flowsheet(self, id_, flowsheet: Union[Dict, str]):
        """Save the flowsheet to the appropriate store.

        Raises:
            ProcessingError, if parsing of JSON failed (see :meth:`DataStoreManager.save()`)
        """
        try:
            self._dsm.save(id_, flowsheet)
        except errors.DatastoreError as err:
            raise errors.ProcessingError(f"While saving flowsheet: {err}")
        except KeyError as err:
            raise errors.ProcessingError(f"While saving flowsheet: {err}")

    def update_flowsheet(self, id_: str) -> Dict:
        """Update flowsheet.

        The returned flowsheet is also saved to the datastore.

        Args:
            id_: Identifier of flowsheet to update.

        Returns:
            Merged value of flowsheets in datastore and current value in memory

        Raises:
            FlowsheetUnknown if the flowsheet id is not known
            FlowsheetNotFound (subclass) if the flowsheet id is known, but it can't be retrieved
            ProcessingError for internal errors
        """
        # Get saved flowsheet from datastore
        try:
            saved = self._load_flowsheet(id_)
        except KeyError:
            raise errors.FlowsheetUnknown(id_)
        except ValueError:
            raise errors.FlowsheetNotFoundInDatastore(id_)
        # Get current value from memory
        try:
            obj = self._get_flowsheet_obj(id_)
        except KeyError:
            raise errors.FlowsheetNotFoundInMemory(id_)
        try:
            obj_dict = self._serialize_flowsheet(id_, obj)
        except ValueError as err:
            raise errors.ProcessingError(f"Cannot serialize flowsheet: {err}")
        # Compare saved and current value
        diff = FlowsheetDiff(saved, obj_dict)
        _log.debug(f"diff: {diff}")
        if not diff:
            # If no difference do nothing
            _log.debug("Stored flowsheet is the same as the flowsheet in memory")
            merged = saved
        else:
            # Otherwise, save this merged value before returning it
            num, pl = len(diff), "s" if len(diff) > 1 else ""
            _log.debug(
                f"Stored flowsheet and model in memory differ by {num} item{pl}"
            )
            self.save_flowsheet(id_, diff.merged())
        # Return [a copy of the] merged value
        return diff.merged(do_copy=True)

    # === Internal methods ===

    def _load_flowsheet(self, id_) -> Union[Dict, str]:
        return self._dsm.load(id_)

    def _get_flowsheet_obj(self, id_):
        """Get a flowsheet with the given ID.
        """
        return self._flowsheets[id_]

    @staticmethod
    def _serialize_flowsheet(id_, flowsheet):
        try:
            result = FlowsheetSerializer(flowsheet, id_).as_dict()
        except (AttributeError, KeyError) as err:
            raise ValueError(f"Error serializing flowsheet: {err}")
        return result

    def _run(self):
        """Run in a separate thread.
        """
        _log.debug(f"Serve forever on localhost:{self._port}")
        try:
            self.serve_forever()
        except Exception as err:
            _log.info(f"Shutting down server due to error: {err}")
            self.shutdown()


class FlowsheetServerHandler(http.server.SimpleHTTPRequestHandler):
    """Handle requests from the IDAES flowsheet visualization (IFV) web page.
    """

    def __init__(self, *args, **kwargs):
        self.directory = None  # silence warning about initialization outside constructor
        super().__init__(*args, **kwargs)
        # Server should return text/javascript MIME type for served JS files (issue 259)
        self.extensions_map[".js"] = "text/javascript"

    # === GET ===

    def do_GET(self):
        """Process a request to receive data.

        Routes:
          * `/app`: Return the web page
          * `/fs`: Retrieve an updated flowsheet.
          * `/path/to/file`: Retrieve file stored static directory
        """
        u, id_ = self._parse_flowsheet_url(self.path)
        _log.debug(f"do_GET: path={self.path} id=={id_}")
        if u.path in ("/app", "/fs") and id_ is None:
            self.send_error(
                400, message=f"Query parameter 'id' is required for '{u.path}'"
            )
            return
        if u.path == "/app":
            self._get_app(id_)
        elif u.path == "/fs":
            self._get_fs(id_)
        else:
            # Try to serve a file
            self.directory = _static_dir  # keep here: overwritten if set earlier
            super().do_GET()

    def _get_app(self, id_):
        """Read index file, process to insert flowsheet identifier, and return it.
        """
        p = Path(_template_dir / "index.html")
        with open(p, "r") as fp:
            s = fp.read()
            page = s.format(flowsheet_id=id_)
        self._write_html(200, page)

    def _get_fs(self, id_: str):
        """Get updated flowsheet.

        Args:
            id_: Flowsheet identifier

        Returns:
            None
        """
        try:
            merged = self.server.update_flowsheet(id_)
        except errors.FlowsheetUnknown as err:
            # User error: user asked for a flowsheet by an unknown ID
            self.send_error(404, message=str(err))
            return
        except (errors.FlowsheetNotFound, errors.ProcessingError) as err:
            # Internal error: flowsheet ID is found, but other things are missing
            self.send_error(500, message=str(err))
            return
        # Return merged flowsheet
        self._write_json(200, merged)

    # === PUT ===

    def do_PUT(self):
        """Process a request to store data.
        """
        u, id_ = self._parse_flowsheet_url(self.path)
        _log.info(f"do_PUT: route={u} id={id_}")
        if u.path in ("/fs",) and id_ is None:
            self.send_error(
                400, message=f"Query parameter 'id' is required for '{u.path}'"
            )
            return
        if u.path == "/fs":
            self._put_fs(id_)

    def _put_fs(self, id_):
        # read  flowsheet from request (read(LENGTH) is required to avoid hanging)
        read_len = int(self.headers.get("Content-Length", "-1"))
        data = utf8_decode(self.rfile.read(read_len))
        # save flowsheet
        try:
            self.server.save_flowsheet(id_, data)
        except errors.ProcessingError as err:
            self.send_error(400, message="Invalid flowsheet", explain=str(err))
            return
        except Exception as err:
            self.send_error(500, message="Unknown error", explain=str(err))
            return
        self.send_response(200, message="success")

    # === Internal methods ===

    def _write_json(self, code, data):
        str_json = json.dumps(data)
        value = utf8_encode(str_json)
        self.send_response(code)
        # self.send_header("Access-Control-Allow-Headers", "Content-Type")
        self.send_header("Content-type", "application/json")
        self.send_header("Content-length", str(len(value)))
        self.end_headers()
        self.wfile.write(value)

    def _write_html(self, code, page):
        value = utf8_encode(page)
        self.send_response(code)
        self.send_header("Content-type", "text/html")
        self.send_header("Content-length", str(len(value)))
        self.end_headers()
        self.wfile.write(value)

    def _parse_flowsheet_url(self, path):
        u, id_ = urlparse(self.path), None
        if u.query:
            queries = dict([q.split("=") for q in u.query.split("&")])
            id_ = queries.get("id", None)
        return u, id_

    # === Logging ===

    def log_message(self, fmt, *args):
        """Override to send messages to our module logger instead of stderr
        """
        msg = "%s - - [%s] %s" % (
            self.address_string(),
            self.log_date_time_string(),
            fmt % args,
        )
        _log.debug(msg)


def utf8_encode(s: str):
    return s.encode(encoding="utf-8")


def utf8_decode(b: bytes):
    return b.decode(encoding="utf-8")


def find_free_port():
    import time

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("127.0.0.1", 0))
    port = s.getsockname()[1]
    s.close()
    time.sleep(1)  # wait for socket cleanup!!!
    return port
